"""Sphinx configuration file."""

# -- Project information -----------------------------------------------------

project = "TopoNetX"
copyright = "2022-2023, PyT-Team, Inc."  # noqa: A001
author = "PyT-Team Authors"
language = "en"

# -- General configuration ---------------------------------------------------

extensions = [
    "nbsphinx",
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_gallery.load_style",
]

templates_path = ["_templates"]
source_suffix = [".rst"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

intersphinx_mapping = {
    "networkx": ("https://networkx.org/documentation/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

# Configure nbsphinx for notebook execution
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]

nbsphinx_execute = "never"
nbsphinx_allow_errors = True

# nbsphinx_thumbnails = {
#     "notebooks/01_simplicial_complexes": "notebooks/sc.png",
#     "notebooks/02_cell_complexes": "notebooks/cc.png",
#     "notebooks/03_combinatorial_complexes": "notebooks/ccc.png",
# }

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_baseurl = "https://pyt-team.github.io/toponetx/"

html_context = {
    "github_user": "pyt-team",
    "github_repo": "TopoNetX",
    "github_version": "main",
    "doc_path": "docs",
}

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/pyt-team/TopoNetX",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
    ],
    "use_edit_page_button": True,
}

html_favicon = "_static/favicon-48.png"

html_show_sourcelink = False

# Exclude copy button from appearing over notebook cell numbers by using :not()
# The default copybutton selector is `div.highlight pre`
# https://github.com/executablebooks/sphinx-copybutton/blob/master/sphinx_copybutton/__init__.py#L82
copybutton_selector = ":not(.prompt) > div.highlight pre"

# -- Options for EPUB output -------------------------------------------------

epub_exclude_files = ["search.html"]



"""Test cell complex class."""

import re

import networkx as nx
import numpy as np
import pytest
import scipy
from scipy.sparse import bmat

from toponetx.classes.cell import Cell
from toponetx.classes.cell_complex import CellComplex

try:
    import hypernetx as hnx
except ImportError:
    hnx = None


class TestCellComplex:
    """Test cell complex class."""

    def test_init_empty_cell_complex(self):
        """Test empty cell complex."""
        CC = CellComplex()
        assert len(CC.cells) == 0
        assert len(CC.nodes) == 0
        assert len(CC.edges) == 0
        assert CC.dim == 0
        assert CC.is_regular

    def test_init_cell_complex_with_list_of_cells(self):
        """Test cell complex with cells."""
        c1 = Cell([1, 2, 3])
        c2 = Cell([1, 2, 3, 4])
        CC = CellComplex([c1, c2])
        assert c1 in CC.cells
        assert c2 in CC.cells
        assert CC.dim == 2

        c1 = Cell((1, 2, 3))
        c2 = Cell((1, 2, 3, 4))
        CC = CellComplex([c1, c2])
        assert len(CC.cells) == 2
        assert len(CC.nodes) == 4
        assert len(CC.edges) == 5

    def test_nodes_and_edges(self):
        """Test cell complex with cells."""
        c1 = Cell([1, 3, 4])
        c2 = Cell([2, 3, 4])
        CC = CellComplex([c1, c2])
        assert set(CC.nodes) == {1, 2, 3, 4}
        assert set(CC.edges) == {(1, 3), (1, 4), (3, 2), (3, 4), (4, 2)}

    def test_init_networkx_graph(self):
        """Test cell complex with networkx graph as input."""
        gr = nx.Graph()
        gr.add_edge(1, 0)
        gr.add_edge(2, 0)
        gr.add_edge(1, 2)
        CC = CellComplex(gr)
        assert CC.dim == 1
        assert len(CC.cells) == 0
        assert len(CC.nodes) == 3
        assert len(CC.edges) == 3

    def test_graph_skeleton(self):
        """Test that graph_skeleton returns nodes/edges with attributes."""
        CC = CellComplex()
        CC.add_node(1, label="a")
        CC.add_node(2, label="b")
        CC.add_edge(1, 2, weight=2.5)
        CC.add_cell((1, 2, 3), rank=2)

        G = CC.graph_skeleton()
        assert set(G.nodes()) == {1, 2, 3}
        assert {tuple(sorted(e)) for e in G.edges()} == {(1, 2), (1, 3), (2, 3)}
        assert G.nodes[1]["label"] == "a"
        assert G.nodes[2]["label"] == "b"
        assert G.edges[1, 2]["weight"] == 2.5

    def test_clone(self):
        """Test CellComplex.clone()."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2, color="blue")
        CC2 = CC.clone()
        assert CC2 is not CC
        assert CC2.cells[(1, 2, 3, 4)] is not CC.cells[(1, 2, 3, 4)]
        CC2.remove_node(1)
        assert 1 in CC.nodes
        assert [1, 2, 3, 4] in CC.cells

    def test_is_regular(self):
        """Test is_regular property."""
        # Test non-regular cell complex
        # allows for constructions of non-regular cells
        CC = CellComplex(regular=False)
        # the "is_regular" method checks if any non-regular cells are added
        assert CC.is_regular
        assert CC.dim == 0

        # test non-regular cell complex
        CC = CellComplex(regular=False)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5, 2, 3, 4, 5], rank=2)  # non-regular 2-cell
        c1 = Cell((1, 2, 3, 4, 5, 1, 2, 3, 4, 5), regular=False)
        CC.add_cell(c1)
        CC.add_cell([5, 6, 7, 8], rank=2)

        assert CC.is_regular is False

    def test_add_cell(self):
        """Test adding cells to a cell complex."""
        # Test adding a single cell
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        assert len(CC.cells) == 1

        # Test adding multiple cells
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        assert len(CC.cells) == 3

        # Test adding cells to CellComplex
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        assert len(CC.cells) == 3
        assert len(CC.nodes) == 8
        assert len(CC.edges) == 10

    def test_add_cells_from(self):
        """Test adding cells from a list of cells or cell lists."""
        # Test adding cells from a list of cells
        CC = CellComplex()
        cells = [Cell((1, 2, 3, 4)), Cell((2, 3, 4, 5))]
        CC.add_cells_from(cells)
        assert len(CC.cells) == 2

        # Test adding cells from a list of cell lists
        CC = CellComplex()
        cell_lists = [[1, 2, 3, 4], [2, 3, 4, 5]]
        CC.add_cells_from(cell_lists, rank=2)
        assert len(CC.cells) == 2

        # Test adding cells from a list of lists
        CC = CellComplex()
        CC.add_cells_from([[1, 2, 4], [1, 2, 7]], rank=2)
        assert len(CC.cells) == 2
        assert len(CC.nodes) == 4
        assert len(CC.edges) == 5

        # Test adding multiple cells to an empty cell complex
        CC = CellComplex()
        CC.add_cells_from([[1, 2, 3], [2, 3, 4]], rank=2)

        # Test adding multiple cells with duplicate vertices to a cell complex
        CC.add_cells_from([[1, 2, 3, 4], [2, 3, 4, 5]], rank=2)
        assert len(CC.cells) == 4

        # Test adding multiple cells with vertices that do not exist in the cell complex
        CC.add_cells_from([[4, 5, 6], [7, 8, 9]], rank=2)

        assert 6 in CC.nodes
        assert 9 in CC.nodes
        assert 8 in CC.nodes

    def test_add_cell_and_remove_cell(self):
        """Test removing one cell and several cells from a cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.remove_cell([1, 2, 3, 4])
        assert len(CC.cells) == 0

        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        CC.remove_cell([1, 2, 3, 4])
        CC.remove_cell([2, 3, 4, 5])
        assert len(CC.cells) == 1

        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        CC.remove_cell([2, 3, 4, 5])
        assert len(CC.cells) == 2
        assert len(CC.nodes) == 8
        assert len(CC.edges) == 10

    def test_incidence_matrix_shape(self):
        """Test the shape of the incidence matrix for the cell complex."""
        CC = CellComplex()
        CC.add_cells_from([[1, 2, 3, 4], [2, 3, 4, 5], [5, 6, 7, 8]], rank=2)
        row_index, col_index, B = CC.incidence_matrix(2, index=True)
        assert B.shape == (10, 3)
        assert len(row_index) == 10
        assert len(col_index) == 3
        row_index, col_index, B = CC.incidence_matrix(1, index=True)
        assert B.shape == (8, 10)
        assert len(row_index) == 8
        assert len(col_index) == 10
        row_index, col_index, B = CC.incidence_matrix(0, index=True)
        assert B.shape == (0, 8)
        assert len(row_index) == 0
        assert len(col_index) == 8

    def test_incidence_matrix_empty_cell_complex(self):
        """Test the incidence matrix for an empty cell complex."""
        CC = CellComplex()
        np.testing.assert_array_equal(
            CC.incidence_matrix(2).toarray(), np.zeros((0, 0))
        )

    def test_incidence_matrix_cell_complex_with_one_cell(self):
        """Test the incidence matrix for a cell complex with only one cell."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        np.testing.assert_array_equal(
            CC.incidence_matrix(2).toarray(), np.array([[1, -1, 1]]).T
        )

    def test_incidence_matrix_cell_complex_with_multiple_cells(self):
        """Test the incidence matrix for a cell complex with multiple cells."""
        CC = CellComplex()
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([1, 3, 4], rank=2)
        np.testing.assert_array_equal(
            CC.incidence_matrix(rank=2).toarray(),
            np.array([[0, 0, 1, -1, 1], [1, -1, 0, 0, 1]]).T,
        )

        # Test non-regular cell complex
        CC = CellComplex(regular=False)
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        np.testing.assert_array_equal(
            CC.incidence_matrix(rank=2).toarray(),
            np.array([[1, -1, 1, 0, 0], [0, 0, 1, -1, 1]]).T,
        )

    def test_incidence_matrix_unsigned_and_signed(self):
        """Test incidence matrix for the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)

        # Test the incidence matrix for the full cell complex
        B2 = CC.incidence_matrix(rank=2, signed=False)
        assert B2.shape == (7, 3)
        assert (B2[:, 0].T.toarray()[0] == np.array([1, 1, 1, 0, 0, 0, 0])).all()
        assert (B2[:, 1].T.toarray()[0] == np.array([0, 0, 1, 1, 1, 0, 0])).all()
        assert (B2[:, 2].T.toarray()[0] == np.array([0, 0, 0, 0, 1, 1, 1])).all()

        # Test the incidence matrix for the full cell complex
        B1 = CC.incidence_matrix(rank=1, signed=False)
        assert B1.shape == (5, 7)
        assert (B1[:, 0].T.toarray()[0] == np.array([1, 1, 0, 0, 0])).all()
        assert (B1[:, 1].T.toarray()[0] == np.array([1, 0, 1, 0, 0])).all()
        assert (B1[:, 2].T.toarray()[0] == np.array([0, 1, 1, 0, 0])).all()

        B1 = CC.incidence_matrix(rank=1, signed=False, index=False)
        assert isinstance(B1, scipy.sparse._csr.csr_matrix)

        B2_signed = CC.incidence_matrix(rank=2, signed=True)
        B1_signed = CC.incidence_matrix(rank=1, signed=True)

        # B1 * B2 == 0
        assert np.sum(B1_signed.dot(B2_signed).toarray()) == 0.0

        CC = CellComplex()
        B1 = CC.incidence_matrix(rank=1)
        assert B1.shape == (0, 0)

        CC = CellComplex()
        B2 = CC.incidence_matrix(rank=2)
        assert B2.shape == (0, 0)

        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        B0 = CC.incidence_matrix(rank=0, signed=False, index=True)
        assert B0[2].shape == (0, 5)
        B0 = CC.incidence_matrix(rank=0, signed=False, index=False)
        assert B0.shape == (0, 5)

    def test_node_to_all_cell_incidence_matrix(self):
        """Test node_to_all_cell_incidence_matrix."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)

        # Test the function without index
        result = CC.node_to_all_cell_incidence_matrix(weight=False, index=False)
        expected_result = scipy.sparse.csc_matrix(
            np.array(
                [
                    [1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                    [1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 1.0, 1.0],
                    [0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 1.0],
                ]
            )
        )
        assert np.allclose(result.toarray(), expected_result.toarray())

        # Test the function with index
        node_index, cell_index, _ = CC.node_to_all_cell_incidence_matrix(
            weight=False, index=True
        )
        expected_node_index = {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
        expected_cell_index = {
            (1, 2): 0,
            (1, 4): 1,
            (2, 3): 2,
            (3, 4): 3,
            (3, 5): 4,
            (4, 5): 5,
            (1, 2, 3, 4): 6,
            (3, 4, 5): 7,
        }
        expected_result = scipy.sparse.csc_matrix(
            np.array(
                [
                    [0.0, 2.0, 1.0, 2.0, 0.0],
                    [2.0, 0.0, 2.0, 1.0, 0.0],
                    [1.0, 2.0, 0.0, 3.0, 2.0],
                    [2.0, 1.0, 3.0, 0.0, 2.0],
                    [0.0, 0.0, 2.0, 2.0, 0.0],
                ]
            )
        )
        assert isinstance(node_index, dict)
        assert isinstance(cell_index, dict)
        assert node_index == expected_node_index
        assert cell_index == expected_cell_index

    def test_node_to_all_cell_adjacnecy_matrix(self):
        """Test node_to_all_cell_adjacnecy_matrix."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)

        # Test the function without index
        result = CC.node_to_all_cell_adjacnecy_matrix(s=2)
        expected_result = np.array(
            [
                [0.0, 2.0, 0.0, 2.0, 0.0],
                [2.0, 0.0, 2.0, 0.0, 0.0],
                [0.0, 2.0, 0.0, 3.0, 2.0],
                [2.0, 0.0, 3.0, 0.0, 2.0],
                [0.0, 0.0, 2.0, 2.0, 0.0],
            ]
        )
        assert np.allclose(result.toarray(), expected_result)

        # Test the function with index
        result = CC.node_to_all_cell_adjacnecy_matrix(index=True)
        expected_node_index = {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
        expected_result = scipy.sparse.csc_matrix(
            np.array(
                [
                    [0.0, 2.0, 1.0, 2.0, 0.0],
                    [2.0, 0.0, 2.0, 1.0, 0.0],
                    [1.0, 2.0, 0.0, 3.0, 2.0],
                    [2.0, 1.0, 3.0, 0.0, 2.0],
                    [0.0, 0.0, 2.0, 2.0, 0.0],
                ]
            )
        )
        assert result[0] == expected_node_index
        assert np.allclose(result[1].toarray(), expected_result.toarray())

    def test_all_cell_to_node_coadjacnecy_matrix(self):
        """Test all cell to node coadjacnecy matrix."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)

        # Test the function without index
        result = CC.all_cell_to_node_coadjacency_matrix()
        expected_result = np.array(
            [
                [0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 2.0, 0.0],
                [1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 2.0, 1.0],
                [1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 2.0, 1.0],
                [0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 2.0, 2.0],
                [0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 2.0],
                [0.0, 1.0, 0.0, 1.0, 1.0, 0.0, 1.0, 2.0],
                [2.0, 2.0, 2.0, 2.0, 1.0, 1.0, 0.0, 2.0],
                [0.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 0.0],
            ]
        )
        assert np.allclose(result.toarray(), expected_result)

        # Test the function with index
        result = CC.all_cell_to_node_coadjacency_matrix(index=True)

        expected_cell_index = {
            (1, 2): 0,
            (1, 4): 1,
            (2, 3): 2,
            (3, 4): 3,
            (3, 5): 4,
            (4, 5): 5,
            (1, 2, 3, 4): 6,
            (3, 4, 5): 7,
        }
        expected_result = np.array(
            [
                [0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 2.0, 0.0],
                [1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 2.0, 1.0],
                [1.0, 0.0, 0.0, 1.0, 1.0, 0.0, 2.0, 1.0],
                [0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 2.0, 2.0],
                [0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 2.0],
                [0.0, 1.0, 0.0, 1.0, 1.0, 0.0, 1.0, 2.0],
                [2.0, 2.0, 2.0, 2.0, 1.0, 1.0, 0.0, 2.0],
                [0.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 0.0],
            ]
        )
        assert isinstance(result, tuple)
        assert isinstance(result[0], dict)
        assert result[0] == expected_cell_index
        assert np.allclose(result[1].toarray(), expected_result)

    def test_from_networkx_graph(self):
        """Test for the from_networkx_graph method."""
        CC = CellComplex()  # Initialize your class object

        # Create a NetworkX graph
        G = nx.Graph()
        G.add_edges_from([(0, 1), (0, 2), (1, 2)])

        # Test the function
        CC.from_networkx_graph(G)
        result = list(CC.edges)
        expected_result = [(0, 1), (0, 2), (1, 2)]
        assert result == expected_result

    def test_euler_characteristic(self):
        """Test euler_characteristic."""
        CC = CellComplex()
        CC.add_cells_from([[1, 2, 3], [1, 2, 4], [2, 3, 4], [1, 2, 3, 4]], rank=2)
        assert CC.euler_characterisitics() == 2

    def test_clear(self):
        """Test the clear method of the cell complex."""
        CC = CellComplex()
        CC.add_cells_from([[1, 2, 3, 4], [5, 6, 7, 8]], rank=2)
        CC.clear()
        assert len(CC.cells) == 0

    def test_add_cell_with_color_feature(self):
        """Test adding a cell with a color feature."""
        CC = CellComplex()
        c1 = Cell((2, 3, 4), color="black")
        CC.add_cell(c1, weight=3)
        CC.add_cell([1, 2, 3, 4], rank=2, color="red")
        CC.add_cell([2, 3, 4, 5], rank=2, color="blue")
        CC.add_cell([5, 6, 7, 8], rank=2, color="green")

        assert CC.cells[(1, 2, 3, 4)]["color"] == "red"
        assert CC.cells[(2, 3, 4, 5)]["color"] == "blue"
        assert CC.cells[(5, 6, 7, 8)]["color"] == "green"

    def test_adjacency_matrix_empty_cell_complex(self):
        """Test adjacency matrix for an empty cell complex."""
        CC = CellComplex()
        np.testing.assert_array_equal(CC.adjacency_matrix(0), np.zeros((0, 0)))

    def test_dirac_operator_matrix(self):
        """Test dirac operator."""
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        m = CC.dirac_operator_matrix()
        size = len(CC.nodes) + len(CC.edges) + len(CC.cells)
        assert m.shape == (size, size)

        L = m.dot(m)

        check_L = bmat(
            [
                [CC.hodge_laplacian_matrix(0), None, None],
                [None, CC.hodge_laplacian_matrix(1), None],
                [None, None, CC.hodge_laplacian_matrix(2)],
            ]
        )

        assert np.linalg.norm((check_L - L).todense()) == 0

        index, m = CC.dirac_operator_matrix(index=True)

        assert 1 in index
        assert 2 in index
        assert 3 in index
        assert 4 in index
        assert len(index) == size

        index, m = CC.dirac_operator_matrix(index=True, signed=False)

        assert np.prod(m.todense() >= 0) == 1

        m = CC.dirac_operator_matrix(signed=False)

        assert np.prod(m.todense() >= 0) == 1

    def test_adjacency_matrix_cell_complex_with_one_cell(self):
        """Test adjacency matrix for a cell complex with one cell."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        A = np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0]])
        np.testing.assert_array_equal(CC.adjacency_matrix(0).todense(), A)

    def test_adjacency_matrix_cell_complex_with_multiple_cell(self):
        """Test adjacency matrix for a cell complex with multiple cells."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([4, 5, 6], rank=2)
        A = np.array(
            [
                [0.0, 1.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 1.0, 0.0, 0.0],
                [1.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 1.0, 0.0],
            ]
        )
        np.testing.assert_array_equal(CC.adjacency_matrix(rank=0).toarray(), A)

    def test_up_laplacian_matrix_empty_cell_complex(self):
        """Test up laplacian matrix for an empty cell complex."""
        CC = CellComplex()
        np.testing.assert_array_equal(CC.up_laplacian_matrix(rank=0), np.zeros((0, 0)))

    def test_up_laplacian_matrix_and_incidence_matrix(self):
        """Test up laplacian matrix and incidence matrix for a cell complex."""
        CC = CellComplex()
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([1, 3, 4], rank=2)

        B1 = CC.incidence_matrix(rank=1)
        L0_up = CC.up_laplacian_matrix(rank=0)
        expected = B1.dot(B1.T)
        np.testing.assert_array_equal(L0_up.toarray(), expected.toarray())

        B2 = CC.incidence_matrix(rank=2)
        L1_up = CC.up_laplacian_matrix(rank=1)
        expected = B2.dot(B2.T)
        np.testing.assert_array_equal(L1_up.toarray(), expected.toarray())

    def test_down_laplacian_matrix_and_incidence_matrix(self):
        """Test down laplacian matrix and incidence matrix for a cell complex."""
        CC = CellComplex()
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([1, 3, 4], rank=2)

        B1 = CC.incidence_matrix(rank=1)
        L1_down = CC.down_laplacian_matrix(rank=1)
        expected = (B1.T).dot(B1)
        np.testing.assert_array_equal(L1_down.toarray(), expected.toarray())

        B2 = CC.incidence_matrix(rank=2)
        L2_down = CC.down_laplacian_matrix(rank=2)
        expected = (B2.T).dot(B2)
        np.testing.assert_array_equal(L2_down.toarray(), expected.toarray())
        CC = CellComplex()
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([1, 3, 4], rank=2)
        with pytest.raises(ValueError):
            CC.down_laplacian_matrix(rank=2, weight="Hello")
        with pytest.raises(ValueError):
            CC.down_laplacian_matrix(rank=3)
        assert np.any(
            CC.down_laplacian_matrix(rank=2, signed=False).todense()
            == np.array([[3.0, 1.0], [1.0, 3.0]])
        )
        row, matrix = CC.down_laplacian_matrix(rank=2, index=True)
        assert row == {(1, 3): 0, (1, 4): 1, (2, 3): 2, (2, 4): 3, (3, 4): 4}
        assert np.any(matrix == np.array([[3.0, 1.0], [1.0, 3.0]]))
        with pytest.raises(ValueError):
            CC.up_laplacian_matrix(rank=2, weight="Hello")
        with pytest.raises(ValueError):
            CC.up_laplacian_matrix(rank=3)
        assert np.any(
            CC.up_laplacian_matrix(rank=1, signed=False).todense()
            == np.array(
                [
                    [1.0, 1.0, 0.0, 0.0, 1.0],
                    [1.0, 1.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 2.0],
                ]
            )
        )
        row, matrix = CC.up_laplacian_matrix(rank=1, index=True)
        assert row == {(1, 3): 0, (1, 4): 1, (2, 3): 2, (2, 4): 3, (3, 4): 4}
        assert np.any(
            matrix
            == np.array(
                [
                    [1.0, 1.0, 0.0, 0.0, 1.0],
                    [1.0, 1.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0, 1.0, 2.0],
                ]
            )
        )
        CC = CellComplex(regular=False)
        CC.add_cell([1, 2, 3, 2], rank=2)
        CC.add_cell([3, 4, 5, 3, 4, 5], rank=2)
        assert CC.incidence_matrix(
            rank=0, index=False, signed=True
        ).todense().shape == (0, 5)
        with pytest.raises(ValueError):
            CC.incidence_matrix(rank=3)
        CC = CellComplex(regular=False)
        CC.add_cell([1, 2, 3, 2], rank=2)
        CC.add_cell([3, 4, 5, 3, 4, 5], rank=2)
        assert np.any(
            CC.incidence_matrix(rank=2).todense()
            == np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 2.0], [0.0, -2.0], [0.0, 2.0]])
        )

    def test_hodge_laplacian_and_up_down_laplacians(self):
        """Test hodge laplacian matrix and up and down laplacians for a cell complex."""
        CC = CellComplex()
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([1, 3, 4], rank=2)

        L1_up = CC.up_laplacian_matrix(rank=1)
        L1_down = CC.down_laplacian_matrix(rank=1)
        L1_hodge = CC.hodge_laplacian_matrix(rank=1)
        expected = L1_up + L1_down
        np.testing.assert_array_equal(L1_hodge.toarray(), expected.toarray())
        CC = CellComplex()
        CC.add_cell([2, 3], rank=1)
        CC.add_cell([1, 3], rank=1)
        assert np.any(
            CC.hodge_laplacian_matrix(rank=1, signed=True).todense()
            == np.array([[2.0, 1.0], [1.0, 2.0]])
        )

    def test_init_empty_abstract_cell(self):
        """Test empty cell complex."""
        cc = CellComplex([], rank=0)
        assert len(cc.nodes) == 0

    def test_init_nonempty_abstract_cell(self):
        """Test non-empty cell complex cells."""
        cc1 = CellComplex((1, 2, 3))
        cc2 = CellComplex(("a", "b", "c", "d"))
        assert len(cc1.nodes) == 3
        assert len(cc2.nodes) == 4

    def test_add_single_cell(self):
        """Test adding single cell to cell complex."""
        cc = CellComplex()
        cc1 = CellComplex()
        cc.add_cell((0, 1, 2, 3, 4), rank=2)
        cc1.add_cell(Cell((0, 1, 2, 3, 4)), rank=2)
        assert len(cc.cells) == 1
        assert cc.shape == (5, 5, 1)  # five nodes, five edges and one 2-cell

    def test_add_many_cells(self):
        """Test adding many cells to cell complex."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3], rank=2)
        cc.add_cell([2, 3, 4, 5], rank=2)
        cc.add_cell([5, 6, 7, 8], rank=2)
        assert len(cc.cells) == 3
        assert cc.shape == (8, 10, 3)

    def test_add_single_node(self):
        """Test adding single node to empty cell complex."""
        cc = CellComplex()
        cc.add_node(1)
        assert len(cc.cells) == 0
        assert len(cc.nodes) == 1

    def test_add_many_nodes(self):
        """Test adding many nodes to empty cell complex."""
        cc = CellComplex()
        cc._add_nodes_from([1, 2, 3])
        assert len(cc.cells) == 0
        assert len(cc.nodes) == 3

    def test_add_node_to_cell(self):
        """Test add node multiple times to cell complex."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3], rank=2)
        cc.add_node(1)
        assert len(cc.cells) == 1
        assert len(cc.nodes) == 3

    def test_remove_cell(self):
        """Test remove cell from cell complex."""
        cc = CellComplex()
        cell = [1, 2, 3]
        cc.add_cell(cell, rank=2)
        cc.remove_cell(cell)
        assert len(cc.cells) == 0
        assert len(cc.nodes) == 3

    def test_remove_many_cells(self):
        """Test remove cells from cell complex."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3, 4], rank=2)
        cc.add_cell([2, 3, 4, 5], rank=2)
        cc.add_cell([5, 6, 7, 8], rank=2)
        cc.remove_cell([1, 2, 3, 4])
        cc.remove_cell([2, 3, 4, 5])
        assert len(cc.cells) == 1

    def test_remove_node(self):
        """Test remove node from cell complex."""
        cc = CellComplex()
        cell = [1, 2, 3]
        cc.add_cell(cell, rank=2)
        cc.remove_node(1)  # removing node removes cells attached to it.
        assert len(cc.cells) == 0
        assert len(cc.edges) == 1
        assert len(cc.nodes) == 2

        with pytest.raises(RuntimeError):
            cc.remove_node(980)

    def test_add_cells_from_list_cells(self):
        """Test adding cells from a list of cells or cell lists."""
        cc = CellComplex()
        cells = [Cell((1, 2, 3, 4)), Cell((2, 3, 4, 5))]
        cc.add_cells_from(cells)
        assert len(cc.cells) == 2

    def test_add_cells_from_list_of_cell_lists(self):
        """Test adding cells from a list of cell lists."""
        cc = CellComplex()
        cell_lists = [[1, 2, 3, 4], [2, 3, 4, 5]]
        cc.add_cells_from(cell_lists, rank=2)
        assert len(cc.cells) == 2

    def test_add_cells_from_list_of_lists(self):
        """Test adding cells from a list of lists."""
        cc = CellComplex()
        cc.add_cells_from([[1, 2, 4], [1, 2, 7]], rank=2)
        assert len(cc.cells) == 2
        assert len(cc.nodes) == 4

    def test_incidence_matrix_empty_abstrcct_cell(self):
        """Test the incidence matrix for an empty cell complex."""
        cc = CellComplex()
        B = cc.incidence_matrix(2)
        np.testing.assert_array_equal(B.toarray(), np.zeros((0, 0)))

    def test_incidence_matrix_abstract_cell_with_one_cell(self):
        """Test the incidence matrix for abstract cell with only one cell."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3], rank=2)
        B = cc.incidence_matrix(2)
        np.testing.assert_array_equal(B.toarray(), np.array([[1, -1, 1]]).T)

    def test_adjacency_matrix_empty_abstrcct_cell(self):
        """Test adjccency matrix for an empty cell complex."""
        cc = CellComplex()
        np.testing.assert_array_equal(cc.adjacency_matrix(0), np.zeros((0, 0)))

    def test_adjccency_matrix_abstrcct_cell_with_one_cell(self):
        """Test adjccency matrix for an abstrcct cell with one cell."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3], rank=2)
        A = np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0]])
        np.testing.assert_array_equal(cc.adjacency_matrix(0).todense(), A)

    def test_coadjccency_matrix_abstrcct_cell_with_one_cell(self):
        """Test coadjccency matrix for an abstrcct cell with one cell."""
        cc = CellComplex()
        cc.add_cell([1, 2, 3], rank=2)
        A = np.array([[0.0, 1.0, 1.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0]])
        np.testing.assert_array_equal(cc.coadjacency_matrix(1).todense(), A)

    def test_repr(self):
        """Test the string representation of the cell complex."""
        CC = CellComplex()
        assert repr(CC) == "CellComplex()"

    def test_len(self):
        """Test the length of the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        assert len(CC) == 3

    def test_iter(self):
        """Test the iterator of the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 4], rank=1)
        assert set(CC) == {
            (1,),
            (2,),
            (3,),
            (4,),
            (2, 1),
            (2, 3),
            (3, 1),
            (2, 4),
            (1, 2, 3),
        }

    def test_cell_equivalence_class(self):
        """Test the cell equivalence class method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 1], rank=2)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 3, 6], rank=2)
        CC.add_cell([3, 4, 1, 2], rank=2)
        CC.add_cell([4, 3, 2, 1], rank=2)
        CC.add_cell([1, 2, 7, 3], rank=2)
        c1 = Cell((1, 2, 3, 4, 5))
        CC.add_cell(c1, rank=2)

        equivalence_classes = CC._cell_equivalence_class()

        assert len(equivalence_classes) == 4

    def test_get_cell_attributes2(self):
        """Test the remove equivalent cells method."""
        import networkx as nx

        G = nx.path_graph(3)

        d = {
            (1, 2, 3, 4): {"color": "red", "attr2": 1},
            (1, 2, 4): {"color": "blue", "attr2": 3},
        }
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell(
            [1, 2, 4],
            rank=2,
        )
        CC.add_cell([3, 4, 8], rank=2)
        CC.set_cell_attributes(d, rank=2)
        assert CC.get_cell_attributes("color", 2) == {
            (1, 2, 3, 4): "red",
            (1, 2, 4): "blue",
        }

    def test_set_cell_attributes(self):
        """Test the set cell attributes method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 4], rank=2)
        CC.add_cell([3, 4, 8], rank=2)
        CC.add_cell([2, 6], rank=2)
        d = {(1, 2, 3, 4): "red", (1, 2, 4): "blue", (2, 6): "green"}
        CC.set_cell_attributes(d, rank=2, name="color")
        assert CC.cells[(2, 6)]["color"] == "green"
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell(
            [1, 2, 4],
            rank=2,
        )
        CC.add_cell([3, 4, 8], rank=2)
        CC.add_cell([4, 5], rank=2)
        d = {
            (1, 2, 3, 4): {"color": "red", "attr2": 1},
            (1, 2, 4): {"color": "blue", "attr2": 3},
            (4, 5): {"color": "green", "attr2": 4},
        }
        CC.set_cell_attributes(d, rank=2)
        assert CC.cells[(1, 2, 3, 4)]["color"] == "red"
        assert CC.cells[(4, 5)]["color"] == "green"
        d = {
            (1, 2, 3, 4): {"color": "red", "attr2": 1},
            (1, 2, 4): {"color": "blue", "attr2": 3},
            (4, 5): {"hello": "world"},
        }
        CC.set_cell_attributes(d, rank=2)

    def test_remove_cells(self):
        """Test remove cells."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell(
            [1, 2, 4],
            rank=2,
        )
        CC.add_cell([3, 4, 8], rank=2)
        CC.remove_cells([(1, 2, 3, 4), (1, 2, 4)])

        assert len(CC.cells) == 1

    def test_degree(self):
        """Test the degree of the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)

        assert CC.degree(1) == 2
        assert CC.degree(2) == 3
        assert CC.degree(3) == 2
        assert CC.degree(4) == 3
        assert CC.degree(5) == 4
        assert CC.degree(6) == 2
        assert CC.degree(7) == 2
        assert CC.degree(8) == 2

        with pytest.raises(NotImplementedError):
            CC.degree(node=1, rank=2)

    def test_cell_neighbors(self):
        """Test the cell neighbours of cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)

        with pytest.raises(NotImplementedError):
            CC.cell_neighbors(1)

    def test_size(self):
        """Test the size of the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)

        c = Cell([1, 2, 3, 4])
        assert CC.size((1, 2, 3, 4)) == 4
        assert CC.size((2, 3, 4, 5)) == 4
        assert CC.size((5, 6, 7, 8)) == 4

        assert CC.size((1, 2, 3, 4), node_set=[1, 2, 3]) == 3
        assert CC.size((2, 3, 4, 5), node_set=[3, 4, 5]) == 3
        assert CC.size((5, 6, 7, 8), node_set=[6, 7, 8]) == 3
        assert CC.size(c, node_set=[1, 2, 3]) == 3
        with pytest.raises(KeyError):
            CC.size((1, 2, 3, 4, 5, 6))

    def test_insert_cell(self):
        """Test inserting a cell into the cell complex."""
        CC = CellComplex()
        CC._insert_cell([1, 2, 3, 4], rank=2)
        CC._insert_cell([2, 3, 4, 5], rank=2)
        CC._insert_cell([5, 6, 7, 8], rank=2)

        assert len(CC.cells) == 3
        assert [cell.elements for cell in CC.cells] == [
            (1, 2, 3, 4),
            (2, 3, 4, 5),
            (5, 6, 7, 8),
        ]

    def test_delete_cell(self):
        """Test deleting a cell from the cell complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)

        CC._delete_cell((1, 2, 3, 4))
        assert len(CC.cells) == 2
        assert [cell.elements for cell in CC.cells] == [(2, 3, 4, 5), (5, 6, 7, 8)]

        CC._delete_cell((2, 3, 4, 5))
        assert len(CC.cells) == 1
        assert [cell.elements for cell in CC.cells] == [(5, 6, 7, 8)]

    def test_shape(self):
        """Test the shape of the cell complex."""
        CC = CellComplex()
        assert CC.shape == (0, 0, 0)

        CC.add_node(1)
        assert CC.shape == (1, 0, 0)

        CC.add_cell([1, 2], rank=1)
        assert CC.shape == (2, 1, 0)

        CC.add_cell([1, 2, 3, 4], rank=2)
        assert CC.shape == (4, 4, 1)

    def test_skeleton(self):
        """Test the skeleton of the cell complex."""
        CC = CellComplex()
        CC.add_node(1)
        CC.add_cell([1, 2], rank=1)
        CC.add_cell([1, 2, 3, 4], rank=2)

        nodes = CC.skeleton(0)
        assert len(nodes) == 4

        edges = CC.skeleton(1)
        assert len(edges) == 4

        cells = CC.skeleton(2)
        assert len(cells) == 1

    def test_filtration(self):
        """Test the filtration of the cell complex."""
        CC = CellComplex([[1, 2, 3], [4, 5]])
        test_filtration = {3: 2, 1: 4, (2, 3): 3.4, (1, 2, 3): 7.6}
        CC.set_filtration(test_filtration, "test")
        assert CC.get_filtration("test") == test_filtration

        with pytest.raises(ValueError):
            test_filtration = {"dummy": "test"}
            CC.set_filtration(test_filtration, "test")

    def test_set_node_attributes(self):
        """Test setting node attributes."""
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        d = {1: {"color": "red", "attr2": 1}, 2: {"color": "blue", "attr2": 3}}
        CC.set_node_attributes(d)

        assert CC.nodes[1]["color"] == "red"
        assert CC.nodes[2]["color"] == "blue"
        assert CC.nodes[1]["attr2"] == 1

    def test_get_node_attributes(self):
        """Test getting node attributes."""
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        d = {1: {"color": "red", "attr2": 1}, 2: {"color": "blue", "attr2": 3}}
        CC.set_node_attributes(d)

        assert CC.get_node_attributes("color") == {1: "red", 2: "blue"}

    def test_set_edge_attributes(self):
        """Test setting edge attributes."""
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        d = {
            (1, 2): {"color": "red", "attr2": 1},
            (2, 3): {"color": "blue", "attr2": 3},
        }
        CC.set_edge_attributes(d)

        assert CC.edges[(1, 2)]["color"] == "red"
        assert CC.edges[(2, 3)]["color"] == "blue"
        assert CC.edges[(1, 2)]["attr2"] == 1

    def test_get_edge_attributes(self):
        """Test getting edge attributes."""
        G = nx.path_graph(3)
        CC = CellComplex(G)
        CC.add_cell([1, 2, 3, 4], rank=2)
        d = {
            (1, 2): {"color": "red", "attr2": 1},
            (2, 3): {"color": "blue", "attr2": 3},
        }
        CC.set_edge_attributes(d)

        assert CC.get_edge_attributes("color") == {(1, 2): "red", (2, 3): "blue"}

    @pytest.mark.skipif(
        hnx is None, reason="Optional dependency 'hypernetx' not installed."
    )
    def test_to_hypergraph(self):
        """Test the conversion of a cell complex to a hypergraph."""
        CC = CellComplex([[1, 2, 3], [4, 5]])
        hx = CC.to_hypergraph()
        assert set(hx.nodes) == set(CC.nodes)
        assert len(hx.edges) == len(CC.edges) + len(CC.cells)

    @pytest.mark.skipif(
        hnx is not None, reason="Optional dependency 'hypernetx' installed."
    )
    def test_to_hypergraph_missing_dependency(self):
        """Test that `to_hypergraph` raises a correct error when `hypernetx` is not installed."""
        CC = CellComplex()
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Cannot transform cell complex to hypergraph, `hypernetx` is not installed."
            ),
        ):
            CC.to_hypergraph()

    def test_to_combinatorial_complex(self):
        """Test the conversion of a cell complex to a combinatorial complex."""
        CC = CellComplex([[1, 2, 3], [4, 5]])
        ccc = CC.to_combinatorial_complex()
        assert len(set(ccc.nodes)) == len(set(CC.nodes))
        assert len(ccc.cells) - len(ccc.nodes) == len(CC.edges) + len(CC.cells)

    def test_to_colored_hypergraph(self):
        """Test the conversion of a cell complex to a colored hypergraph."""
        CC = CellComplex([[1, 2, 3], [4, 5]])
        chg = CC.to_colored_hypergraph()
        assert len(set(chg.nodes)) == len(set(CC.nodes))
        assert len(chg.cells) == len(CC.edges) + len(CC.cells)

    def test_restrict_to_nodes(self):
        """Test restricting a cell complex to a subset of nodes."""
        CC = CellComplex([[1, 2, 3], [3, 4, 5], [1, 4]])
        CC.set_filtration({(1, 2, 3): 1, (3, 4, 5): 2, (1, 4): 0, 3: 1}, "test")
        restricted = CC.restrict_to_nodes({1, 2, 3, 4})
        assert (1, 2, 3) in restricted.cells
        assert (1, 4) in restricted.edges
        assert (3, 4) in restricted.edges
        assert (3, 4, 5) not in restricted.cells
        assert (4, 5) not in restricted.edges
        assert restricted.get_filtration("test") == {(1, 2, 3): 1, (1, 4): 0, 3: 1}

    def test_get_cell_attributes(self):
        """Unit test for the get_cell_attributes method."""
        CC = CellComplex()
        d = {
            (1, 2, 3, 4): {"color": "red", "attr2": 1},
            (1, 2, 4): {"color": "blue", "attr2": 3},
        }
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell(
            [1, 2, 4],
            rank=2,
        )
        CC.add_cell([3, 4, 8], rank=2)
        CC.set_cell_attributes(d, 2)

        attributes = CC.get_cell_attributes("color", 2)

        with pytest.raises(ValueError):
            CC.set_cell_attributes(d, 4)

        # should not throw error
        CC.set_cell_attributes({(5, 6, 7): {"color": "red"}}, 2)

        attributes = CC.get_cell_attributes("color", 2)

        with pytest.raises(ValueError):
            attributes = CC.get_cell_attributes("color", 4)

        assert attributes == {(1, 2, 3, 4): "red", (1, 2, 4): "blue"}

    def test_remove_equivalent_cells(self):
        """Unit test for the remove_equivalent_cells method."""
        CC = CellComplex()

        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 1], rank=2)
        CC.add_cell([1, 2, 4], rank=2)
        CC.add_cell([3, 4, 8], rank=2)
        assert len(CC.cells) == 4
        CC.remove_equivalent_cells()
        assert len(CC.cells) == 3

    def test_is_insertable_cycle(self):
        """Unit test for the is_insertable_cycle method."""
        CC = CellComplex()

        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)

        assert CC.is_insertable_cycle([1, 2, 3, 4])
        assert not CC.is_insertable_cycle([1, 2, 3, 4, 1, 2])
        assert not CC.is_insertable_cycle([1, 2, 1])

        with pytest.warns():
            CC.is_insertable_cycle([6, 7, 8], warnings_dis=True)

    def test_incidence_matrix(self):
        """Unit test for the incidence_matrix method."""
        CC = CellComplex()

        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        B1 = CC.incidence_matrix(1)
        B2 = CC.incidence_matrix(2)
        assert B1.dot(B2).todense().tolist() == [
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
        ]

    def test_restrict_to_cells(self):
        """Test restricting a cell complex to a subset of cells and edges."""
        CC = CellComplex([[1, 2, 3, 4], [3, 4, 5], [1, 2, 3, 6]])
        CC.set_filtration(
            {(1, 2, 3, 4): 1, (1, 2): 2, (3, 5): 3, (4, 5): 4, (3, 4, 5): 5}, "test"
        )
        restricted = CC.restrict_to_cells(
            {(1, 2, 3, 4), (3, 5), CC.cells.raw((1, 2, 3, 6))}
        )
        assert (1, 2, 3, 4) in restricted.cells
        assert (1, 2, 3, 6) in restricted.cells
        assert (3, 4, 5) not in restricted.cells
        assert (3, 5) in restricted.edges
        assert (4, 5) not in restricted.edges
        assert restricted.get_filtration("test") == {
            (1, 2, 3, 4): 1,
            (2, 1): 2,
            (3, 5): 3,
        }

    def test_is_connected(self):
        """Test is connected."""
        # connected cell complex
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4, 5], rank=2)
        CC.add_cell([1, 2, 7, 8], rank=2)
        assert CC.is_connected() is True

        # disconnected cell complex
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4, 5], rank=2)
        CC.add_cell([7, 8, 9, 10], rank=2)
        assert CC.is_connected() is False

    def test_singletons(self):
        """Test singletons."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        CC.add_node(0)
        CC.add_node(10)
        assert CC.singletons() == [0, 10]

    def test_remove_singletons(self):
        """Test remove singletons."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)
        CC.add_cell([5, 6, 7, 8], rank=2)
        CC.add_node(0)
        CC.add_node(10)
        CC.remove_singletons()
        assert CC.singletons() == []

    def test_contains(self):
        """Test __contains__."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)

        assert 1 in CC
        assert 2 in CC
        assert 3 in CC
        assert 4 in CC
        assert 5 in CC
        assert 10 not in CC
        assert (1, 2) in CC
        assert (10, 12) not in CC
        assert (1, 2, 3, 4) in CC
        assert (1, 2, 3, 4, 5) not in CC

    def test_neighbors(self):
        """Test neighbors of nodes."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([2, 3, 4, 5], rank=2)

        assert list(CC.neighbors(1)) == [2, 4]

    def test_insert_cell_2(self):
        """Test for the _insert_cell method."""
        CC = CellComplex()
        CC._insert_cell([1, 2, 3], color="red")
        assert len(CC._cells._cells) == 1
        CC._insert_cell((1, 2, 3), attr1=5)
        assert len(CC._cells._cells) == 1
        CC._insert_cell(Cell([1, 2, 3], name="Cell 1"))
        assert len(CC._cells._cells) == 1
        CC._insert_cell([1, 2, 4])
        assert len(CC._cells._cells) == 2

    def test_insert_cell_invalid_input(self):
        """Test _insert_cell method with invalid input."""
        CC = CellComplex()
        with pytest.raises(TypeError):
            CC._insert_cell("invalid_input")
        with pytest.raises(TypeError):
            CC._insert_cell(123)
        with pytest.raises(TypeError):
            CC._insert_cell({"elements": [1, 2, 3]})

    def test_delete_cell_2(self):
        """Test for the _delete_cell method."""
        CC = CellComplex()
        CC._insert_cell([1, 2, 3], color="red")
        CC._insert_cell([4, 5, 6], color="blue")
        CC._delete_cell((1, 2, 3))
        assert len(CC._cells._cells) == 1
        assert len(CC.cells) == 1
        CC._delete_cell((4, 5, 6))
        assert len(CC._cells._cells) == 0
        assert len(CC.cells) == 0

    def test_delete_cell_invalid_input(self):
        """Test for the _delete_cell method with invalid input."""
        CC = CellComplex()
        with pytest.raises(KeyError):
            CC._delete_cell((1, 2, 3))

    def test_number_of_nodes(self):
        """Test for the number_of_nodes method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_node(4)
        CC.add_node(5)
        assert CC.number_of_nodes() == 5
        assert CC.number_of_nodes([1, 2, 3]) == 3
        assert CC.number_of_nodes([1, 6]) == 1

    def test_number_of_edges(self):
        """Test for the number_of_edges method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_node(4)
        CC.add_node(5)
        assert CC.number_of_edges() == 3
        assert CC.number_of_edges([(1, 2), (2, 3)]) == 2
        assert CC.number_of_edges([(4, 5), (6, 7)]) == 0

    def test_number_of_cells(self):
        """Test for the number_of_cells method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_node(4)
        CC.add_node(5)
        assert CC.number_of_cells() == 1
        assert CC.number_of_cells([(4, 5), (6, 7)]) == 0

    def test_order(self):
        """Test for the order method."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_node(4)
        CC.add_node(5)
        assert CC.order() == 5

    def test_neighbors_error(self):
        """Test to check for ValueError in neighbors method."""
        CC = CellComplex()
        with pytest.raises(KeyError):
            CC.neighbors(10)

    def test_get_cell_data(self):
        """Test the get_cell_data method of CellComplex."""
        CC = CellComplex()

        CC.add_node("A", attribute_name="Value A")
        CC.add_node("B", attribute_name="Value B")

        CC.add_edge("A", "B", attribute_name="Value AB")

        CC.add_cell(["A", "B", "C"], rank=2, attribute_name="Value C")

        data = CC.get_cell_data("A", 0, "attribute_name")
        assert data == "Value A"

        data = CC.get_cell_data(("A", "B"), 1, "attribute_name")
        assert data == "Value AB"

        data = CC.get_cell_data("B", 0)
        assert data == {"attribute_name": "Value B"}

        with pytest.raises(KeyError):
            CC.get_cell_data("D", 1, "attribute_name")

        with pytest.raises(KeyError):
            CC.get_cell_data("A", 0, "invalid_attribute")

        with pytest.raises(ValueError):
            CC.get_cell_data(["C"], 2, "invalid_attribute")

        with pytest.raises(KeyError):
            CC.get_cell_data(["D", "F"], 2, "invalid_attribute")

        with pytest.raises(KeyError):
            CC.get_cell_data("C", 2, "invalid_attribute")

        with pytest.raises(KeyError):
            CC.get_cell_data("E", 0)

        with pytest.raises(ValueError):
            CC.get_cell_data("A", 3, "attribute_name")

    def test_set_cell_data(self):
        """Test the set_cell_data method of CellComplex."""
        CC = CellComplex()

        CC.add_node("A")
        CC.add_node("B")

        CC.add_edge("A", "B")

        CC.add_cell(["A", "B", "C"], rank=2)

        CC.set_cell_data("A", 0, "attribute_name", "Value A")
        assert CC.nodes["A"]["attribute_name"] == "Value A"

        CC.set_cell_data(("A", "B"), 1, "attribute_name", "Value AB")
        assert CC.edges[("A", "B")]["attribute_name"] == "Value AB"

        CC.set_cell_data(["A", "B", "C"], 2, "attribute_name", "Value C")
        assert CC.cells[("A", "B", "C")]["attribute_name"] == "Value C"

        with pytest.raises(KeyError):
            CC.set_cell_data("D", 0, "attribute_name", "Value D")

        with pytest.raises(KeyError):
            CC.set_cell_data("D", 1, "attribute_name", "Value D")

        with pytest.raises(KeyError):
            CC.set_cell_data(["D", "E"], 2, "attribute_name", "Value D")

        with pytest.raises(ValueError):
            CC.set_cell_data("A", 3, "attribute_name", "Value A")

    def test_get_cell_data_after_set(self):
        """Test the get_cell_data method after setting cell data."""
        CC = CellComplex()

        CC.add_node("A")

        CC.set_cell_data("A", 0, "attribute_name", "Value A")

        data = CC.get_cell_data("A", 0, "attribute_name")
        assert data == "Value A"

    def test_get_cell_data_without_attr(self):
        """Test the get_cell_data method without specifying an attribute name."""
        CC = CellComplex()

        CC.add_node("A")
        CC.add_node("B", attribute_name="Value B")

        data = CC.get_cell_data("A", 0)
        assert len(data) == 0

    def test_hodge_laplacian_matrix(self):
        """Test the hodge_laplacian_matrix method of CellComplex."""
        CC = CellComplex()

        # Add cells to the complex for testing
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        CC.add_cell([4, 5, 6], rank=2)

        # Test case 1: Rank is 0
        rank = 0
        signed = True
        weight = None
        index = False

        result = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        assert result.shape == (6, 6)

        assert result.shape == (6, 6)

        # Test case 2: Rank is 0 with index=True
        index = True

        result, index_list = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        assert len(result) == 6

        assert len(result) == 6

        # Test case 3: Rank is 1 and maxdim is 2
        rank = 1

        result = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        # Test case 4: Rank is 1 and maxdim is 2 with index=True
        index = True

        index_list, result = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        # Assert the index_list is of type list
        assert isinstance(index_list, dict)

        # Test case 5: Rank is 2 and maxdim is 2
        rank = 2

        result = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        # Test case 6: Rank is 2 and maxdim is 2 with index=True
        index = True

        index_list, result = CC.hodge_laplacian_matrix(rank, signed, weight, index)

        # Assert the index_list is of type list
        assert isinstance(index_list, dict)

        # Test case 7: Rank is 2 and maxdim is not 2
        rank = 3

        with pytest.raises(ValueError):
            CC.hodge_laplacian_matrix(rank, signed, weight, index)

    def test_to_hasse_graph(self):
        """Test to hasse graph function."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        G = CC.to_hasse_graph()
        assert len(G.nodes) == 9
        assert len(G.edges) == 12
        assert (1,) in G.nodes
        assert (2,) in G.nodes
        assert (3,) in G.nodes
        assert (4,) in G.nodes
        assert (1, 2, 3, 4) in G.nodes

    def test_init_type_exception(self):
        """Test if incorrect datatype raises a TypeError exception."""
        with pytest.raises(TypeError):
            CellComplex(cells=1)

    def test_dim_edgecases(self):
        """Test if dim is testing for edge cases."""
        CC = CellComplex()
        CC.add_node(1)
        assert CC.dim == 0

    def test_skeleton_method_exception(self):
        """Test if the skeleton method raises an exception."""
        with pytest.raises(ValueError):
            CC = CellComplex()
            CC.skeleton(rank=4)

        with pytest.raises(ValueError):
            CC = CellComplex()
            CC.skeleton(rank=3)

    def test_getitem_dunder_method(self):
        """Test if the dunder __getitem__ method returns the appropriate neighbors of the given node."""
        CC = CellComplex()
        CC.add_node(1, color="red")
        CC.add_edge(1, 2, weight=10)
        CC.add_cell((2, 3, 4), rank=2, weight=5)

        assert CC[1]["color"] == "red"
        assert CC[(1,)]["color"] == "red"
        assert CC[(1, 2)]["weight"] == 10
        assert CC[(2, 3, 4)]["weight"] == 5

        with pytest.raises(KeyError):
            _ = CC[5]

        CC[(1, 2)]["new"] = 1
        assert CC[(1, 2)]["new"] == 1

    def test_remove_nodes(self):
        """Test remove nodes method of the class Cell Complex."""
        cc = CellComplex()
        cell = [1, 2, 3]
        cc.add_cell(cell, rank=2)
        cc.remove_nodes([1, 2, 3])
        assert len(cc.cells) == 0
        assert len(cc.edges) == 0
        assert len(cc.nodes) == 0

    def test_add_cell_regularity_conditions(self):
        """Test regularity conditions for add cell method."""
        CC = CellComplex()
        with pytest.raises(RuntimeError) as exp_exception:
            CC.add_cell((1, 1, 2, 3), rank=0)

        assert (
            str(exp_exception.value)
            == "Use `add_node` to insert nodes or zero ranked cells."
        )

        with pytest.raises(ValueError) as exp_exception:
            CC.add_cell((1, 1, 2, 3), rank=1)

        assert (
            str(exp_exception.value)
            == "rank 1 cells (edges) must have exactly two nodes"
        )

        with pytest.raises(ValueError) as exp_exception:
            CC.add_cell((1, 1), rank=1)

        assert (
            str(exp_exception.value)
            == " invalid insertion : self-loops are not allowed."
        )

        with pytest.raises(ValueError) as exp_exception:
            CC.add_cell((1, 1, 2, 3), rank=2)

        assert (
            str(exp_exception.value)
            == "Invalid cycle condition for cell [1, 1, 2, 3]. This input cell is not inserted, check if cell is irregular or if boundary edges of the cell are in the complex."
        )

        with pytest.raises(ValueError) as exp_exception:
            CC.add_cell(1, rank=2)

        assert str(exp_exception.value) == "invalid input, input 1 must be iterable"

        with pytest.raises(ValueError) as exp_exception:
            CC.add_cell({1, 5, 2, 3}, rank=4)

        assert (
            str(exp_exception.value)
            == "Add cell only supports adding cells of dimensions 0,1 or 2-- got 4"
        )

    def test_get_linegraph(self):
        """Test for the get_linegraph function of Cell Complex."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        with pytest.raises(ValueError):
            CC.get_linegraph((1, 2, 3, 4, 5, 6))
        assert len(list(CC.get_linegraph(1).adjacency())) == 10
        with pytest.raises(TypeError):
            CC.get_linegraph(cells=2)

    def test_coadjacencies_edge_cases(self):
        """Test coadjacency and adjacency function edge cases."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        ind, incidence = CC.adjacency_matrix(rank=1, index=True)
        assert ind == {(1, 2): 0, (1, 4): 1, (2, 3): 2, (3, 4): 3}
        assert np.any(
            incidence.todense()
            == np.array(
                [
                    [0.0, 1.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0, 1.0],
                    [1.0, 1.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0, 0.0],
                ]
            )
        )
        ind, incidence = CC.coadjacency_matrix(1, index=True)
        assert ind == {(1, 2): 0, (1, 4): 1, (2, 3): 2, (3, 4): 3}
        assert np.any(
            incidence.todense()
            == np.array(
                [
                    [0.0, 1.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0, 1.0],
                    [1.0, 1.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0, 0.0],
                ]
            )
        )

    def test_hodge_laplacian_edgecases(self):
        """Test hodge laplacian edge cases."""
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        ind, laplacian_matrix = CC.hodge_laplacian_matrix(0, index=True, signed=False)
        assert ind == {1: 0, 2: 1, 3: 2, 4: 3, 5: 4}
        assert np.any(
            laplacian_matrix.todense()
            == np.array(
                [
                    [2.0, 1.0, 0.0, 1.0, 0.0],
                    [1.0, 2.0, 1.0, 0.0, 0.0],
                    [0.0, 1.0, 3.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0, 3.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 2.0],
                ]
            )
        )
        ind = CC.hodge_laplacian_matrix(0, index=False, signed=False)
        assert np.any(
            ind.todense()
            == np.array(
                [
                    [2.0, 1.0, 0.0, 1.0, 0.0],
                    [1.0, 2.0, 1.0, 0.0, 0.0],
                    [0.0, 1.0, 3.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0, 3.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 2.0],
                ]
            )
        )
        CC = CellComplex()
        CC.add_cell([6, 8], rank=1)
        CC.add_cell([1, 2, 3], rank=2)
        assert CC.hodge_laplacian_matrix(2).todense() == np.array([[3]])
        CC = CellComplex()
        CC.add_node(1)
        CC.add_edge(1, 2)
        with pytest.raises(ValueError):
            CC.hodge_laplacian_matrix(rank=2)
        CC = CellComplex()
        CC.add_cell([1, 2, 3], rank=2)
        assert np.any(
            CC.hodge_laplacian_matrix(rank=2, signed=False).todense() == np.array([[3]])
        )
        CC = CellComplex()
        CC.add_cell([1, 2, 3, 4], rank=2)
        CC.add_cell([3, 4, 5], rank=2)
        np.any(
            CC.hodge_laplacian_matrix(1, signed=False).todense()
            == np.array(
                [
                    [3.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 3.0, 1.0, 0.0, 0.0, 1.0],
                    [0.0, 1.0, 3.0, 0.0, 1.0, 0.0],
                    [1.0, 0.0, 0.0, 4.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 3.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0, 3.0],
                ]
            )
        )



"""Tests for the `simplex_trie` module."""

import pytest

from toponetx.classes.simplex import Simplex
from toponetx.classes.simplex_trie import SimplexNode, SimplexTrie


class TestSimplexNode:
    """Tests for the `SimplexNode` class."""

    def test_invalid_root(self) -> None:
        """Test that invalid root nodes are rejected."""
        with pytest.raises(ValueError):
            _ = SimplexNode(label="invalid", parent=None)

    def test_node_len(self) -> None:
        """Test the length of the node."""
        root_node: SimplexNode[int] = SimplexNode(label=None)
        assert len(root_node) == 0

        child_node: SimplexNode[int] = SimplexNode(label=1, parent=root_node)
        assert len(child_node) == 1

        grandchild_node: SimplexNode[int] = SimplexNode(label=2, parent=child_node)
        assert len(grandchild_node) == 2

    def test_node_repr(self) -> None:
        """Test the string representation of the node."""
        root_node = SimplexNode(label=None)
        assert repr(root_node) == "SimplexNode(None, None)"

        child_node = SimplexNode(label=1, parent=root_node)
        assert repr(child_node) == "SimplexNode(1, SimplexNode(None, None))"

    def test_simplex_property(self) -> None:
        """Test the `simplex` property of a node."""
        root_node = SimplexNode(label=None)
        assert root_node.simplex is None

        child_node = SimplexNode(label=1, parent=root_node)
        assert child_node.simplex == Simplex((1,))

        grandchild_node = SimplexNode(label=2, parent=child_node)
        assert grandchild_node.simplex == Simplex((1, 2))


class TestSimplexTrie:
    """Tests for the `SimplexTree` class."""

    def test_insert(self) -> None:
        """Test that the internal data structures of the simplex trie are correct after insertion."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))

        assert trie.shape == [3, 3, 1]

        assert set(trie.root.children.keys()) == {1, 2, 3}
        assert set(trie.root.children[1].children.keys()) == {2, 3}
        assert set(trie.root.children[1].children[2].children.keys()) == {3}
        assert set(trie.root.children[2].children.keys()) == {3}

        # the label list should contain the nodes of each depth according to their label
        label_to_simplex = {
            1: {1: [(1,)], 2: [(2,)], 3: [(3,)]},
            2: {2: [(1, 2)], 3: [(1, 3), (2, 3)]},
            3: {3: [(1, 2, 3)]},
        }

        assert len(trie.label_lists) == len(label_to_simplex)
        for depth, label_list in trie.label_lists.items():
            assert depth in label_to_simplex
            assert len(label_list) == len(label_to_simplex[depth])
            for label, nodes in label_list.items():
                assert len(nodes) == len(label_to_simplex[depth][label])
                for node, expected in zip(
                    nodes, label_to_simplex[depth][label], strict=True
                ):
                    assert node.simplex.elements == expected

    def test_len(self) -> None:
        """Test the `__len__` method of a simplex trie."""
        trie = SimplexTrie()
        assert len(trie) == 0

        trie.insert((1, 2, 3))
        assert len(trie) == 7

    def test_getitem(self) -> None:
        """Test the `__getitem__` method of a simplex trie."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))

        assert trie[(1,)].simplex == Simplex((1,))
        assert trie[(1, 2)].simplex == Simplex((1, 2))
        assert trie[(1, 2, 3)].simplex == Simplex((1, 2, 3))

        with pytest.raises(KeyError):
            _ = trie[(0,)]

    def test_iter(self) -> None:
        """Test the iteration of the trie."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))
        trie.insert((2, 3, 4))
        trie.insert((0, 1))

        # We guarantee a specific ordering of the simplices when iterating. Hence, we explicitly compare lists here.
        assert [node.simplex.elements for node in trie] == [
            (0,),
            (1,),
            (2,),
            (3,),
            (4,),
            (0, 1),
            (1, 2),
            (1, 3),
            (2, 3),
            (2, 4),
            (3, 4),
            (1, 2, 3),
            (2, 3, 4),
        ]

    def test_cofaces(self) -> None:
        """Test the cofaces method."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))
        trie.insert((1, 2, 4))

        # no ordering is guaranteed for the cofaces method
        assert {node.simplex.elements for node in trie.cofaces((1,))} == {
            (1,),
            (1, 2),
            (1, 3),
            (1, 4),
            (1, 2, 3),
            (1, 2, 4),
        }
        assert {node.simplex.elements for node in trie.cofaces((2,))} == {
            (2,),
            (1, 2),
            (2, 3),
            (2, 4),
            (1, 2, 3),
            (1, 2, 4),
        }

    def test_is_maximal(self) -> None:
        """Test the `is_maximal` method."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))
        trie.insert((1, 2, 4))

        assert trie.is_maximal((1, 2, 3))
        assert trie.is_maximal((1, 2, 4))
        assert not trie.is_maximal((1, 2))
        assert not trie.is_maximal((1, 3))
        assert not trie.is_maximal((1, 4))
        assert not trie.is_maximal((2, 3))

        with pytest.raises(ValueError):
            trie.is_maximal((5,))

    def test_skeleton(self) -> None:
        """Test the skeleton method."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))
        trie.insert((1, 2, 4))

        # no ordering is guaranteed for the skeleton method
        assert {node.simplex.elements for node in trie.skeleton(0)} == {
            (1,),
            (2,),
            (3,),
            (4,),
        }
        assert {node.simplex.elements for node in trie.skeleton(1)} == {
            (1, 2),
            (1, 3),
            (1, 4),
            (2, 3),
            (2, 4),
        }
        assert {node.simplex.elements for node in trie.skeleton(2)} == {
            (1, 2, 3),
            (1, 2, 4),
        }

        with pytest.raises(ValueError):
            _ = next(trie.skeleton(-1))
        with pytest.raises(ValueError):
            _ = next(trie.skeleton(3))

    def test_remove_simplex(self) -> None:
        """Test the removal of simplices from the trie."""
        trie = SimplexTrie()
        trie.insert((1, 2, 3))

        trie.remove_simplex((1, 2))
        assert len(trie) == 5



"""Test mesh dataset."""

import pytest

from toponetx.datasets.mesh import coseg, shrec_16, stanford_bunny


class TestMeshDatasets:
    """Test datasets utils."""

    def test_stanford_bunny(self):
        """Test stanford_bunny."""
        simplicialbunny = stanford_bunny("simplicial")

        assert len(simplicialbunny) == 2503

        cellbunny = stanford_bunny("cell")

        assert len(cellbunny) == 2503

        with pytest.raises(ValueError):
            stanford_bunny("polyhedral")

    def test_shrec_16_small(self):
        """Test shrec_16."""
        shrec_training, shrec_testing = shrec_16(size="small")

        assert len(shrec_training["complexes"]) == 100
        assert len(shrec_testing["complexes"]) == 20
        assert len(shrec_training["label"]) == 100
        assert len(shrec_testing["label"]) == 20
        assert len(shrec_training["node_feat"]) == 100
        assert len(shrec_testing["node_feat"]) == 20
        assert len(shrec_training["edge_feat"]) == 100
        assert len(shrec_testing["edge_feat"]) == 20
        assert len(shrec_training["face_feat"]) == 100
        assert len(shrec_testing["face_feat"]) == 20

    def test_shrec_16_full(self):
        """Test shrec_16."""
        shrec_training, shrec_testing = shrec_16(size="full")

        assert len(shrec_training["complexes"]) == 480
        assert len(shrec_testing["complexes"]) == 120
        assert len(shrec_training["label"]) == 480
        assert len(shrec_testing["label"]) == 120
        assert len(shrec_training["node_feat"]) == 480
        assert len(shrec_testing["node_feat"]) == 120
        assert len(shrec_training["edge_feat"]) == 480
        assert len(shrec_testing["edge_feat"]) == 120
        assert len(shrec_training["face_feat"]) == 480
        assert len(shrec_testing["face_feat"]) == 120

    def test_shrec_16_invalid(self):
        """Test shrec_16."""
        with pytest.raises(ValueError):
            shrec_16(size="huge")

    def test_coseg(self):
        """Test coseg."""
        coseg_data = coseg(data="alien")

        assert len(coseg_data["complexes"]) != 0
        assert len(coseg_data["node_feat"]) != 0
        assert len(coseg_data["edge_feat"]) != 0
        assert len(coseg_data["face_feat"]) != 0
        assert len(coseg_data["face_label"]) != 0

        nodes, _, faces = coseg_data["complexes"][0].shape
        n_nodes = coseg_data["node_feat"][0].shape
        n_faces = coseg_data["face_feat"][0].shape
        assert nodes == n_nodes[0]
        assert faces == n_faces[0]
        assert len(coseg_data["face_label"][0]) == faces

        with pytest.raises(ValueError):
            coseg(data="cars")



"""Test Cell class."""

from collections.abc import Iterable

import pytest

from toponetx.classes.cell import Cell


class TestCell:
    """Test Cell class."""

    def test_create_cell_with_less_than_2_edges(self):
        """Test creating a cell with less than 2 edges."""
        elements = [1]
        with pytest.raises(ValueError):
            Cell(elements)

    def test_create_cell_with_self_loops(self):
        """Test creating a cell with self-loops."""
        elements = [1, 2, 3, 2]
        with pytest.raises(ValueError):
            Cell(elements)

    def test_sign_with_invalid_input(self):
        """Test signing with invalid input."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        with pytest.raises(TypeError):
            cell.sign(123)
        with pytest.raises(ValueError):
            cell.sign((1,))
        with pytest.raises(KeyError):
            cell.sign((1, 4))

    def test_get_item_with_invalid_key(self):
        """Test getting an item with an invalid key."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        with pytest.raises(KeyError):
            cell["invalid_key"]

    def test_are_homotopic_with_invalid_input(self):
        """Test checking if cells are homotopic with invalid input."""
        elements = [1, 2, 3]
        cell1 = Cell(elements)
        with pytest.raises(TypeError):
            Cell._are_homotopic(cell1, 123)

    def test_elements_and_nodes(self):
        """Test elements attributes of a cell."""
        cell = Cell([1, 2, 3])
        assert cell.elements == (1, 2, 3)

    def test_is_regular(self):
        """Test is_regular property."""
        cell = Cell([1, 2, 3])
        assert cell.is_regular
        cell = Cell([1, 2, 3, 2], regular=False)
        assert not cell.is_regular

    def test_reverse(self):
        """Test reverse."""
        cell = Cell([1, 2, 3])
        cell_reverse = cell.reverse()
        expected = Cell([3, 2, 1])
        assert cell_reverse.elements == expected.elements

    def test_self_loop_irregular(self):
        """Test loop irregular."""
        with pytest.raises(ValueError):
            Cell([1, 1, 2], regular=False)

    def test___setitem__(self):
        """Test __setitem__."""
        c = Cell([1, 2, 3])
        c["weight"] = 1
        assert c["weight"] == 1

    def test___iter__(self):
        """Test __iter__."""
        c = Cell([1, 2, 3])
        for _ in c:
            pass

    def test___repr__(self):
        """Test __repr__."""
        c = Cell([1, 2, 3])
        print(c)

    def test_is_homotopic_to(self):
        """Test is_homotopic_to."""
        c = Cell([1, 2, 3])

        assert c.is_homotopic_to(Cell([1, 2, 3]))
        assert not c.is_homotopic_to(Cell([2, 4, 1]))

    def test_are_homotopic(self):
        """Test _are_homotopic."""
        c1 = Cell([1, 2, 3])
        c2 = Cell([2, 3, 1])

        assert Cell._are_homotopic(c1, c2)
        assert Cell._are_homotopic(c1, (1, 2, 3))

        with pytest.raises(TypeError):
            assert Cell._are_homotopic((1, 2, 3), c1)

    def test_get_item(self):
        """Test the __getitem__ method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        cell["name"] = "cell1"
        assert cell["name"] == "cell1"
        with pytest.raises(KeyError):
            _ = cell["invalid_key"]

    def test_set_item(self):
        """Test the __setitem__ method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        cell["name"] = "cell1"
        assert cell["name"] == "cell1"

    def test_is_regular_property(self):
        """Test the is_regular property of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert cell.is_regular

    def test_len(self):
        """Test the __len__ method of Cell."""
        elements = [1, 2, 3, 4]
        cell = Cell(elements)
        assert len(cell) == 4

    def test_iter(self):
        """Test the __iter__ method of Cell."""
        elements = [1, 2, 3, 4]
        cell = Cell(elements)
        assert isinstance(iter(cell), Iterable)
        assert list(cell) == elements

    def test_sign(self):
        """Test the sign method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert cell.sign((1, 2)) == 1
        assert cell.sign((2, 1)) == -1
        with pytest.raises(KeyError):
            cell.sign((2, 4))
        with pytest.raises(ValueError):
            cell.sign((1,))

    def test_contains(self):
        """Test the __contains__ method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert 1 in cell
        assert 4 not in cell

    def test_repr(self):
        """Test the __repr__ method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert repr(cell) == f"Cell({cell.elements})"

    def test_boundary(self):
        """Test the boundary property of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert list(cell.boundary) == [(1, 2), (2, 3), (3, 1)]

    def test_elements(self):
        """Test the elements property of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        assert cell.elements == tuple(elements)

    def test_reverse2(self):
        """Test the reverse method of Cell."""
        elements = [1, 2, 3]
        cell = Cell(elements)
        reversed_cell = cell.reverse()
        assert reversed_cell.elements == tuple(elements[::-1])

    def test_valid_regular_cell(self):
        """Test a valid regular cell."""
        elements = [1, 2, 3]
        assert Cell.is_valid_cell(elements, regular=True)

    def test_valid_irregular_cell(self):
        """Test a valid irregular cell."""
        elements = [1, 2, 3, 4, 1, 2, 3]
        assert Cell.is_valid_cell(elements, regular=False)

    def test_invalid_single_element_cell(self):
        """Test an invalid cell with a single element."""
        elements = [1]
        assert not Cell.is_valid_cell(elements, regular=True)

    def test_invalid_repeated_element_cell(self):
        """Test an invalid cell with repeated elements."""
        elements = [1, 1, 1, 1]
        assert not Cell.is_valid_cell(elements, regular=False)

    def test_invalid_regular_cell(self):
        """Test an invalid regular cell."""
        elements = [1, 2, 3, 2]
        assert not Cell.is_valid_cell(elements, regular=True)

    def test_invalid_irregular_cell(self):
        """Test an invalid irregular cell."""
        elements = [1, 2, 2, 4]
        assert not Cell.is_valid_cell(elements, regular=False)



"""Class for creation and manipulation of simplicial complexes.

The class also supports attaching arbitrary attributes and data to cells.
"""

from collections.abc import Collection, Generator, Hashable, Iterable, Iterator
from itertools import combinations
from typing import Any, Self

import networkx as nx
import numpy as np
from scipy.sparse import csr_matrix, dok_matrix

from toponetx.classes.complex import Complex
from toponetx.classes.reportviews import NodeView, SimplexView
from toponetx.classes.simplex import Simplex
from toponetx.classes.simplex_trie import SimplexTrie

try:
    from gudhi import SimplexTree
except ImportError:
    SimplexTree = None

try:
    from hypernetx import Hypergraph
except ImportError:
    Hypergraph = None

__all__ = ["SimplicialComplex"]


# TODO: `ElementType` also needs to be bound to SupportsLessThanT but that is not accessible?
class SimplicialComplex[ElementType: Hashable](Complex):
    """Class representing a simplicial complex.

    Class for construction boundary operators, Hodge Laplacians,
    higher order (co)adjacency operators from a collection of
    simplices.

    A simplicial complex is a topological space of a specific kind, constructed by
    "gluing together" points, line segments, triangles, and their higher-dimensional
    counterparts. It is a generalization of the notion of a triangle in a triangulated surface,
    or a tetrahedron in a tetrahedralized 3-dimensional manifold. Simplicial complexes are the
    basic objects of study in combinatorial topology.

    For example, a triangle is a simplicial complex because it is a collection of three
    points that are connected to each other in a specific way. Similarly, a tetrahedron is a
    simplicial complex because it is a collection of four points that are connected to each
    other in a specific way. These simplices can be thought of as the "building blocks" of a
    simplicial complex, and the complex itself is constructed by combining these building blocks
    in a specific way. For example, a 2-dimensional simplicial complex could be a collection of
    triangles that are connected to each other to form a surface, while a 3-dimensional simplicial
    complex could be a collection of tetrahedra that are connected to each other to form a solid object.

    The SimplicialComplex class is a class for representing simplicial complexes,
    which are a type of topological space constructed by "gluing together" points, line segments,
    triangles, and higher-dimensional counterparts. The class provides methods for computing boundary
    operators, Hodge Laplacians, and higher-order adjacency operators on the simplicial complex.
    It also allows for compatibility with NetworkX and the GUDHI library.

    Features:

    1. The SimplicialComplex class allows for the dynamic construction of simplicial complexes,
        enabling users to add or remove simplices from the complex after its initial creation.
    2. The class provides methods for computing boundary operators, Hodge Laplacians,
        and higher-order adjacency operators on the simplicial complex.
    3. The class is compatible with the gudhi library, allowing users to leverage the powerful
        algorithms and data structures provided by this package.
    4. The class supports the attachment of arbitrary attributes and data to simplices,
        enabling users to store and manipulate additional information about these objects.
    5. The class has robust error handling and input validation, ensuring reliable and easy use of the class.

    Parameters
    ----------
    simplices : iterable, optional
        Iterable of maximal simplices that define the simplicial complex.
    **kwargs : keyword arguments, optional
        Attributes to add to the complex as key=value pairs.

    Attributes
    ----------
    complex : dict
        A dictionary that can be used to store additional information about the complex.

    Notes
    -----
    A simplicial complex is determined by its maximal simplices, simplices that are not
    contained in any other simplices. If a maximal simplex is inserted, all faces of this
    simplex will be inserted automatically.

    Examples
    --------
    Define a simplicial complex using a set of maximal simplices:

    >>> SC = tnx.SimplicialComplex([[1, 2, 3], [2, 3, 5], [0, 1]])

    TopoNetX is also compatible with NetworkX, allowing users to create a simplicial complex from a NetworkX graph.
    Existing node and edge attributes are copied to the simplicial complex:

    >>> G = nx.Graph()
    >>> G.add_edge(0, 1, weight=4)
    >>> G.add_edges_from([(0, 3), (0, 4), (1, 4)])
    >>> SC = tnx.SimplicialComplex(simplices=G)
    >>> SC.add_simplex([1, 2, 3])
    >>> SC.simplices
    SimplexView([(0,), (1,), (3,), (4,), (2,), (0, 1), (0, 3), (0, 4), (1, 4), (1, 2), (1, 3), (2, 3), (1, 2, 3)])
    >>> SC[(0, 1)]["weight"]
    4
    """

    _simplex_trie: SimplexTrie[ElementType]

    def __init__(self, simplices=None, **kwargs) -> None:
        super().__init__(**kwargs)

        self._simplex_trie = SimplexTrie()

        if isinstance(simplices, nx.Graph):
            _simplices: dict[tuple[Hashable, ...], Any] = {}
            for simplex, data in simplices.nodes(data=True):
                _simplices[(simplex,)] = data
            for u, v, data in simplices.edges(data=True):
                _simplices[(u, v)] = data

            simplices = []
            for simplex, data in _simplices.items():
                simplices.append(Simplex(simplex, **data))
            self.add_simplices_from(simplices)
        elif isinstance(simplices, Iterable):
            self.add_simplices_from(simplices)
        elif simplices is not None:
            raise TypeError(
                f"Input simplices must be given as Iterable, got {type(simplices)}."
            )

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of simplicial complex, i.e., the number of simplices for each dimension.

        Returns
        -------
        tuple of ints
            This gives the number of cells in each rank.

        Examples
        --------
        >>> SC = SimplicialComplex([[0, 1], [1, 2, 3], [2, 3, 4]])
        >>> SC.shape
        (5, 5, 2)
        """
        return tuple(self._simplex_trie.shape)

    @property
    def dim(self) -> int:
        """Maximal dimension of the simplices in this complex.

        Returns
        -------
        int
            The dimension of the simplicial complex.

        Examples
        --------
        >>> SC = SimplicialComplex([[0, 1], [1, 2, 3], [2, 3, 4]])
        >>> SC.dim
        2
        """
        return len(self._simplex_trie.shape) - 1

    @property
    def nodes(self) -> NodeView:
        """Return the list of nodes in the simplicial complex.

        Returns
        -------
        NodeView
            A NodeView object representing the nodes of the simplicial complex.
        """
        return NodeView(
            {
                frozenset(node.elements): node.attributes
                for node in self._simplex_trie.skeleton(0)
            },
            Simplex,
        )

    @property
    def simplices(self) -> SimplexView:
        """Set of all simplices in the simplicial complex.

        Returns
        -------
        SimplexView
            A SimplexView object representing the set of all simplices in the simplicial complex.
        """
        return SimplexView(self._simplex_trie)

    def is_maximal(self, simplex: Iterable[ElementType]) -> bool:
        """Check if simplex is maximal, i.e., not a face of any other simplex in the complex.

        Parameters
        ----------
        simplex : Iterable
            The Simplex to check.

        Returns
        -------
        bool
            True if the given simplex is maximal in this simplicial complex, False
            otherwise.

        Raises
        ------
        ValueError
            If simplex is not in the simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([[1, 2, 3]])
        >>> SC.is_maximal([1, 2, 3])
        True
        >>> SC.is_maximal([1, 2])
        False
        """
        return self._simplex_trie.is_maximal(tuple(sorted(simplex)))

    def get_maximal_simplices_of_simplex(
        self, simplex: Iterable[Hashable]
    ) -> set[Simplex[ElementType]]:
        """Get maximal simplices of simplex.

        Parameters
        ----------
        simplex : Iterable
            The simplex for which to compute the maximal simplices.

        Returns
        -------
        set of Simplex
            Set of maximal simplices of the given simplex.
        """
        return {
            node.simplex
            for node in self._simplex_trie.cofaces(tuple(sorted(simplex)))
            if self._simplex_trie.is_maximal(node.elements)
        }

    def skeleton(self, rank: int) -> list[Simplex[ElementType]]:
        """Compute the rank-skeleton of the simplicial complex containing simplices of given rank.

        Parameters
        ----------
        rank : int
            The rank of the skeleton to compute.

        Returns
        -------
        list of Simplex
            Simplices of rank `rank` in the simplicial complex.
        """
        return sorted(node.simplex for node in self._simplex_trie.skeleton(rank))

    def __str__(self) -> str:
        """Return a detailed string representation of the simplicial complex.

        Returns
        -------
        str
            A string representation containing information about the shape and dimension
            of the simplicial complex.
        """
        return f"Simplicial Complex with shape {self.shape} and dimension {self.dim}"

    def __repr__(self) -> str:
        """Return a string representation of the simplicial complex.

        Returns
        -------
        str
            A string representation of the simplicial complex.
        """
        return "SimplicialComplex()"

    def __len__(self) -> int:
        """Compute number of simplices.

        Returns
        -------
        int
            Number of vertices in the complex.
        """
        return self.shape[0]

    def __getitem__(self, simplex: Iterable[ElementType]) -> dict[Hashable, Any]:
        """Get the data associated with the given simplex.

        Parameters
        ----------
        simplex : Any
            The simplex to retrieve.

        Returns
        -------
        dict[Hashable, Any]
            The data associated with the given simplex.

        Raises
        ------
        KeyError
            If the simplex is not present in the simplicial complex.
        """
        if isinstance(simplex, Hashable) and not isinstance(simplex, Iterable):
            simplex = (simplex,)
        return self._simplex_trie[tuple(sorted(simplex))].attributes

    def __iter__(self) -> Iterator[Simplex[ElementType]]:
        """Iterate over all simplices (faces) of the simplicial complex.

        Returns
        -------
        Iterator[frozenset[Hashable]]
            An iterator over all simplices in the simplicial complex.

        Raises
        ------
        KeyError
            If the simplex is not present in the simplicial complex.
        """
        return iter(self.simplices)

    def __contains__(self, atom: Any) -> bool:
        """Check whether this simplicial complex contains the given atom.

        Parameters
        ----------
        atom : Any
            The atom to be checked.

        Returns
        -------
        bool
            Returns `True` if this simplicial complex contains the atom, else `False`.
        """
        if not isinstance(atom, Iterable):
            return (atom,) in self._simplex_trie
        return tuple(sorted(atom)) in self._simplex_trie

    @staticmethod
    def get_boundaries(
        simplices: Iterable, min_dim=None, max_dim=None
    ) -> set[frozenset]:
        """Get boundaries of the given simplices.

        Parameters
        ----------
        simplices : Iterable
            Iterable of simplices for which to compute the boundaries.
        min_dim : int, optional
            Constrain the max dimension of faces.
        max_dim : int, optional
            Constrain the max dimension of faces.

        Returns
        -------
        set of frozensets
            Set of simplices that are boundaries of the given simplices. If `min_dim` or `max_dim` are given, only
            simplices with dimension in the given range are returned.
        """
        if not isinstance(simplices, Iterable):
            raise TypeError(
                f"Input simplices must be given as a list or tuple, got {type(simplices)}."
            )

        face_set = set()
        for simplex in simplices:
            start = (
                min(max_dim + 1, len(simplex)) if max_dim is not None else len(simplex)
            )
            end = min_dim if min_dim is not None else 0

            for r in range(start, end, -1):
                for face in combinations(simplex, r):
                    face_set.add(frozenset(face))

        return face_set

    def remove_maximal_simplex(self, simplex: Collection) -> None:
        """Remove maximal simplex from simplicial complex.

        Parameters
        ----------
        simplex : Collection
            The simplex to be removed from the simplicial complex.

        Raises
        ------
        KeyError
            If simplex is not in simplicial complex.
        ValueError
            If simplex is not maximal.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex((1, 2, 3, 4), weight=1)
        >>> SC.add_simplex((1, 2, 3, 4, 5))
        >>> SC.remove_maximal_simplex((1, 2, 3, 4, 5))
        """
        if isinstance(simplex, Hashable) and not isinstance(simplex, Iterable):
            simplex_ = (simplex,)
        elif isinstance(simplex, Iterable):
            simplex_ = sorted(simplex)
        else:
            raise TypeError("`simplex` must be Hashable or Iterable.")

        node = self._simplex_trie.find(simplex_)
        if node is None:
            raise ValueError(
                f"Simplex {simplex_} is not a part of the simplicial complex."
            )
        if len(node.children) != 0:
            raise ValueError(f"Simplex {simplex_} is not maximal.")

        self._simplex_trie.remove_simplex(simplex_)

    def remove_nodes(self, node_set: Iterable[Hashable]) -> None:
        """Remove the given nodes from the simplicial complex.

        Any simplices that become invalid due to the removal of nodes are also removed.

        Parameters
        ----------
        node_set : Iterable
            The nodes to be removed from the simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([(1, 2), (2, 3), (3, 4)])
        >>> SC.remove_nodes([1])
        >>> SC.simplices
        SimplexView([(2,), (3,), (4,), (2, 3), (3, 4)])
        """
        for node in node_set:
            self._simplex_trie.remove_simplex((node,))

    def add_node(self, node: Hashable, **kwargs) -> None:
        """Add node to simplicial complex.

        Parameters
        ----------
        node : Hashable or Simplex
            The node to be added to the simplicial complex.
        **kwargs : keyword arguments, optional
            Additional attributes to be associated with the node.
        """
        if not isinstance(node, Hashable):
            raise TypeError(f"Input node must be Hashable, got {type(node)} instead.")

        if isinstance(node, Simplex):
            if len(node) != 1:
                raise ValueError(
                    f"Input node must be a singleton Simplex, got {node} instead."
                )
            self.add_simplex(node, **kwargs)
        else:
            self.add_simplex([node], **kwargs)

    def add_simplex(
        self, simplex: Iterable[ElementType] | ElementType, **kwargs
    ) -> None:
        """Add simplex to simplicial complex.

        In case sub-simplices are missing, they are added without attributes to the
        simplicial complex to fulfill the simplex requirements.

        If the given simplex is already part of the simplicial complex, its attributes
        will be updated by the provided ones.

        Parameters
        ----------
        simplex : Collection
            The simplex to be added to the simplicial complex.

            If a `Simplex` object is given, its attributes will be copied to the
            simplicial complex. `kwargs` take precedence over the attributes of the
            `Simplex` object.
        **kwargs : keyword arguments, optional
            Additional attributes to be associated with the simplex.
        """
        Simplex.validate_attributes(kwargs)

        # Support some short-hand calls for adding nodes. The user does not have to
        # provide a single-element list but can give the node directly.
        if isinstance(simplex, Hashable) and not isinstance(simplex, Iterable):
            simplex = [simplex]
        if isinstance(simplex, str):
            simplex = [simplex]

        if not isinstance(simplex, Iterable) and not isinstance(simplex, Simplex):
            raise TypeError(
                f"Input simplex must be Iterable or Simplex, got {type(simplex)} instead."
            )

        if isinstance(simplex, Simplex):
            simplex_ = simplex.elements
            kwargs = simplex._attributes | kwargs
        else:
            simplex_ = frozenset(simplex)
            if len(simplex_) != len(simplex):
                raise ValueError("a simplex cannot contain duplicate nodes")

        self._simplex_trie.insert(sorted(simplex_), **kwargs)

    def add_simplices_from(self, simplices) -> None:
        """Add simplices from iterable to simplicial complex.

        Parameters
        ----------
        simplices : iterable
            Iterable of simplices to be added to the simplicial complex.
        """
        # for simplex in simplices:
        #     if isinstance(simplex, Hashable) and not isinstance(simplex, Iterable):
        #         simplex = [simplex]
        #     if isinstance(simplex, str):
        #         simplex = [simplex]
        #
        #     if not isinstance(simplex, Iterable) and not isinstance(simplex, Simplex):
        #         raise TypeError(
        #             f"Input simplex must be Iterable or Simplex, got {type(simplex)} instead."
        #         )
        #
        #     if isinstance(simplex, Simplex):
        #         simplex_ = simplex.elements
        #         attr = simplex._properties
        #     else:
        #         simplex_ = frozenset(simplex)
        #         if len(simplex_) != len(simplex):
        #             raise ValueError("a simplex cannot contain duplicate nodes")
        #         attr = {}
        #
        #     self._simplex_trie.insert_raw(sorted(simplex_), **attr)
        # self._simplex_trie.restore_simplex_property()
        for s in simplices:
            self.add_simplex(s)

    def get_cofaces(
        self, simplex: Iterable[ElementType], codimension: int
    ) -> list[Simplex[ElementType]]:
        """Get cofaces of simplex.

        Parameters
        ----------
        simplex : Iterable of AtomType
            The simplex for which to compute the cofaces.
        codimension : int
            The codimension of the returned cofaces. If zero, all cofaces are returned.

        Returns
        -------
        list of tuples
            The cofaces of the given simplex.
        """
        simplex = tuple(sorted(simplex))
        # TODO: This can be optimized inside the simplex trie.
        cofaces = self._simplex_trie.cofaces(simplex)
        return [
            coface.simplex
            for coface in cofaces
            if len(coface) - len(simplex) >= codimension
        ]

    def get_star(self, simplex: Iterable[ElementType]) -> list[Simplex[ElementType]]:
        """Get star.

        Notes
        -----
        This function is equivalent to calling `get_cofaces(simplex, 0)`.

        Parameters
        ----------
        simplex : Iterable of AtomType
            The simplex for which to compute the star.

        Returns
        -------
        list of tuples
            The star of the given simplex.
        """
        return self.get_cofaces(simplex, 0)

    def set_simplex_attributes(self, values, name: str | None = None) -> None:
        """Set simplex attributes.

        Parameters
        ----------
        values : dict
            Either provide a mapping from simplices to values or a dict of dicts. In the former case, the attribute
            `name` for each simplex is set to the corresponding value. In the latter case, the entire dictionary
            is used to update the attributes of the simplices.
        name : str, optional
            The name of the attribute to set.

        Notes
        -----
        If the dict contains simplices that are not in `self.simplices`, they are
        silently ignored.

        Examples
        --------
        After computing some property of the simplex of a simplicial complex, you may want
        to assign a simplex attribute to store the value of that property for
        each simplex:

        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> d = {(1, 2, 3): "red", (1, 2, 4): "blue"}
        >>> SC.set_simplex_attributes(d, name="color")
        >>> SC[(1, 2, 3)]["color"]
        'red'

        If you provide a dictionary of dictionaries as the second argument,
        the entire dictionary will be used to update simplex attributes:

        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 3, 4])
        >>> SC.add_simplex([1, 2, 3])
        >>> SC.add_simplex([1, 2, 4])
        >>> d = {
        ...     (1, 3, 4): {"color": "red", "attr2": 1},
        ...     (1, 2, 4): {"color": "blue", "attr2": 3},
        ... }
        >>> SC.set_simplex_attributes(d)
        >>> SC[(1, 3, 4)]["color"]
        'red'
        """
        if name is not None:
            # if `values` is a dict using `.items()` => {simplex: value}
            for simplex, value in values.items():
                if simplex in self.simplices:
                    self[simplex][name] = value
        else:
            for simplex, d in values.items():
                if simplex in self.simplices:
                    self[simplex].update(d)

    def get_node_attributes(self, name: str) -> dict[Hashable, Any]:
        """Get node attributes from simplicial complex.

        Parameters
        ----------
        name : str
           Attribute name.

        Returns
        -------
        dict
            Dictionary mapping each node to the value of the given attribute name.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> SC.set_simplex_attributes({1: "red", 2: "blue", 3: "black"}, name="color")
        >>> SC.get_node_attributes("color")
        {1: 'red', 2: 'blue', 3: 'black'}
        """
        return {
            node.label: node.attributes[name]
            for node in self._simplex_trie.skeleton(0)
            if name in node.attributes
        }

    def get_simplex_attributes(
        self, name: str, rank: int | None = None
    ) -> dict[tuple[ElementType, ...], Any]:
        """Get node attributes from simplical complex.

        Parameters
        ----------
        name : str
           Attribute name.
        rank : int, optional
            Restrict the returned attribute values to simplices of a specific rank.

        Returns
        -------
        dict[tuple[Hashable, ...], Any]
            Dictionary mapping each simplex to the value of the given attribute name.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> d = {(1, 2): "red", (2, 3): "blue", (3, 4): "black"}
        >>> SC.set_simplex_attributes(d, name="color")
        >>> SC.get_simplex_attributes("color")
        {(1, 2): 'red', (2, 3): 'blue', (3, 4): 'black'}
        """
        if rank is None:
            return {
                n.elements: self.simplices[n][name]
                for n in self.simplices
                if name in self.simplices[n]
            }
        return {
            n.elements: self.simplices[n][name]
            for n in self.skeleton(rank)
            if name in self[n]
        }

    @staticmethod
    def get_edges_from_matrix(matrix):
        """Get edges from matrix.

        Parameters
        ----------
        matrix : numpy or scipy array
            The matrix to get the edges from.

        Returns
        -------
        list of int
            List of indices where the operator is not zero.

        Notes
        -----
        Most operaters (e.g. adjacencies/(co)boundary maps) that describe
        connectivity of the simplicial complex
        can be described as a G whose nodes are the simplices used to
        construct the operator and whose edges correspond to the entries
        in the matrix where the operator is not zero.

        This property implies that many computations on simplicial complexes
        can be reduced to G computations.
        """
        rows, cols = np.where(np.sign(np.abs(matrix)) > 0)
        return zip(rows.tolist(), cols.tolist(), strict=True)

    def incidence_matrix(
        self, rank, signed: bool = True, weight: str | None = None, index: bool = False
    ) -> csr_matrix | tuple[dict, dict, csr_matrix]:
        """Compute incidence matrix of the simplicial complex.

        Getting the matrix that correpodnds to the boundary matrix of the input SC.

        Parameters
        ----------
        rank : int
            For which rank of simplices to compute the incidence matrix.
        signed : bool, default=True
            Whether to return the signed or unsigned incidence matrix.
        weight : str, optional
            The name of the simplex attribute to use as weights for the incidence
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the incidence matrix.

        Returns
        -------
        row_indices, col_indices : dict
            Dictionary assigning each row and column of the incidence matrix to a
            simplex. Only returned if `index` is True.
        incidence_matrix : scipy.sparse.csr.csr_matrix
            The incidence matrix.

        Raises
        ------
        ValueError
            If rank is negative or larger than the dimension of the simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> B1 = SC.incidence_matrix(1)
        >>> B2 = SC.incidence_matrix(2)
        """
        if rank < 0:
            raise ValueError(f"Rank must be non-negative, got {rank}.")
        if rank > self.dim:
            raise ValueError(
                f"Rank cannot be larger than the dimension of the complex, got {rank}."
            )

        # TODO: Get rid of this special case...
        if rank == 0:
            boundary = dok_matrix((1, self._simplex_trie.shape[rank]), dtype=np.float32)
            boundary[0, 0 : self._simplex_trie.shape[rank]] = 1

            if index:
                simplex_dict_d = {
                    simplex.elements: i for i, simplex in enumerate(self.skeleton(0))
                }
                return {}, simplex_dict_d, boundary.tocsr()
            return boundary.tocsr()
        idx_simplices, idx_faces, values = [], [], []

        simplex_dict_d = {
            tuple(sorted(simplex)): i for i, simplex in enumerate(self.skeleton(rank))
        }
        simplex_dict_d_minus_1 = {
            tuple(sorted(simplex)): i
            for i, simplex in enumerate(self.skeleton(rank - 1))
        }
        for simplex, idx_simplex in simplex_dict_d.items():
            for i, left_out in enumerate(np.sort(list(simplex))):
                idx_simplices.append(idx_simplex)
                values.append((-1) ** i)
                face = frozenset(simplex).difference({left_out})
                idx_faces.append(simplex_dict_d_minus_1[tuple(sorted(face))])

        if len(values) != (rank + 1) * len(simplex_dict_d):
            raise RuntimeError("Error in computing the incidence matrix.")

        boundary = csr_matrix(
            (values, (idx_faces, idx_simplices)),
            dtype=np.float32,
            shape=(
                len(simplex_dict_d_minus_1),
                len(simplex_dict_d),
            ),
        )

        if not signed:
            boundary = abs(boundary)
        if index:
            return simplex_dict_d_minus_1, simplex_dict_d, boundary
        return boundary

    def coincidence_matrix(
        self, rank, signed: bool = True, weight=None, index: bool = False
    ):
        """Compute coincidence matrix of the simplicial complex.

        This is also called the coboundary matrix.

        Parameters
        ----------
        rank : int
            For which rank of simplices to compute the coincidence matrix.
        signed : bool, default=True
            Whether to return the signed or unsigned coincidence matrix.
        weight : str, optional
            The name of the simplex attribute to use as weights for the coincidence
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the coincidence
            matrix.

        Returns
        -------
        row_indices, col_indices : dict
            Dictionary assigning each row and column of the coincidence matrix to a
            simplex. Only returned if `index` is True.
        coincidence_matrix : scipy.sparse.csr.csr_matrix
            The coincidence matrix.
        """
        if index:
            idx_faces, idx_simplices, boundary = self.incidence_matrix(
                rank, signed=signed, weight=weight, index=True
            )
            return idx_faces, idx_simplices, boundary.T

        return self.incidence_matrix(rank, signed=signed, weight=weight, index=False).T

    def hodge_laplacian_matrix(
        self, rank: int, signed: bool = True, weight=None, index: bool = False
    ):
        """Compute hodge-laplacian matrix for the simplicial complex.

        Parameters
        ----------
        rank : int
            Dimension of the Laplacian matrix.
        signed : bool, default=True
            Whether to return the signed or unsigned hodge laplacian.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Indicates whether to return the indices that define the hodge laplacian
            matrix.

        Returns
        -------
        index : list
            List assigning each row and column of the laplacian matrix to a simplex.
            Only available when `index` is True.
        laplacian : scipy.sparse.csr.csr_matrix
            The hodge laplacian matrix of rank `rank`.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> L1 = SC.hodge_laplacian_matrix(1)
        """
        if rank == 0:
            row, column, B_next = self.incidence_matrix(
                rank + 1, weight=weight, index=True
            )
            L_hodge = B_next @ B_next.transpose()
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return row, L_hodge
            return L_hodge
        if rank < self.dim:
            row, column, B_next = self.incidence_matrix(
                rank + 1, weight=weight, index=True
            )
            row, column, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_hodge = B_next @ B_next.transpose() + B.transpose() @ B
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return column, L_hodge
            return L_hodge
        if rank == self.dim:
            row, column, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_hodge = B.transpose() @ B
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return column, L_hodge
            return L_hodge

        raise ValueError(
            f"Rank should be larger than 0 and <= {self.dim} (maximal dimension simplices), got {rank}"
        )

    def dirac_operator_matrix(
        self,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ):
        """Compute dirac operator matrix matrix.

        Parameters
        ----------
        signed : bool, default=False
            Whether the returned dirac matrix should be signed (i.e., respect
            orientations) or unsigned.
        weight : str, optional
            The name of the simplex attribute to use as weights for the dirac operator
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the dirac operator matrix.

        Returns
        -------
        row_indices, col_indices : dict
            List identifying rows and columns of the dirac operator matrix. Only
            returned if `index` is True.
        scipy.sparse.csr.csc_matrix
            The dirac operator matrix of this simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> SC.dirac_operator_matrix()
        """
        from scipy.sparse import bmat

        index_set = []
        incidence = {}
        for i in range(self.dim + 1):
            _, indexi, Bi = self.incidence_matrix(i, weight=weight, index=True)
            index_set.append(indexi)
            incidence[(i, i + 1)] = Bi
        dirac = []
        for i in range(self.dim + 1):
            row = []
            for j in range(self.dim + 1):
                if (i, j) in incidence:
                    row.append(incidence[(i + 1, j + 1)])
                elif (j, i) in incidence:
                    row.append(incidence[(j + 1, i + 1)].T)
                else:
                    row.append(None)
            dirac.append(row)

        dirac_mat = bmat(dirac)
        if index:
            d = {}
            shift = 0
            for i in index_set:
                i = {k: v + shift for k, v in i.items()}
                d.update(i)
                shift = len(d)

            if signed:
                return d, dirac_mat
            return d, abs(dirac_mat)

        if signed:
            return dirac_mat
        return abs(dirac_mat)

    def normalized_laplacian_matrix(self, rank: int, weight: str | None = None):
        """Return the normalized hodge Laplacian matrix of simplicial complex .

        The normalized hodge Laplacian is the matrix

        .. math::
            N_d = D^{-1/2} L_d D^{-1/2}

        where `L` is the simplicial complex Laplacian and `D` is the diagonal matrix of
        simplices of rank d.

        Parameters
        ----------
        rank : int
            Rank of the hodge laplacian matrix.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.

        Returns
        -------
        Scipy sparse matrix
            The normalized hodge Laplacian matrix.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([[1, 2, 3], [2, 3, 5], [0, 1]])
        >>> SC.normalized_laplacian_matrix(1)
        """
        import scipy as sp

        L_hodge = self.hodge_laplacian_matrix(rank)
        m, n = L_hodge.shape
        diags_ = abs(L_hodge).sum(axis=1)

        with np.errstate(divide="ignore"):
            diags_sqrt = 1.0 / np.sqrt(diags_)
        diags_sqrt[np.isinf(diags_sqrt)] = 0
        diags_sqrt = sp.sparse.csr_array(
            sp.sparse.spdiags(diags_sqrt.T, 0, m, n, format="csr")
        )

        return sp.sparse.csr_matrix(diags_sqrt @ (L_hodge @ diags_sqrt))

    def up_laplacian_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ):
        """Compute the up Laplacian matrix of the simplicial complex.

        Parameters
        ----------
        rank : int
            Rank of the up Laplacian matrix.
        signed : bool
            Whether to return the signed or unsigned up laplacian.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the laplacian.

        Returns
        -------
        row_indices, col_indices : list
            List identifying the rows and columns of the laplacian matrix with the
            simplices of this complex. Only returned if `index` is True.
        up_laplacian : scipy.sparse.csr.csr_matrix
            The upper laplacian matrix of this simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([[1, 2, 3], [2, 3, 5], [0, 1]])
        >>> SC.up_laplacian_matrix(1)
        """
        if weight is not None:
            raise ValueError("`weight` is not supported in this version")

        if rank < self.dim and rank >= 0:
            row, _, B_next = self.incidence_matrix(rank + 1, weight=weight, index=True)
            L_up = B_next @ B_next.transpose()
        else:
            raise ValueError(
                f"Rank should larger than 0 and <= {self.dim - 1} (maximal dimension cells-1), got {rank}"
            )
        if not signed:
            L_up = abs(L_up)

        if index:
            return row, L_up
        return L_up

    def down_laplacian_matrix(
        self, rank, signed: bool = True, weight=None, index: bool = False
    ):
        """Compute the down Laplacian matrix of the simplicial complex.

        Parameters
        ----------
        rank : int
            Rank of the down Laplacian matrix.
        signed : bool
            Whether to return the signed or unsigned down laplacian.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the laplacian.

        Returns
        -------
        indices : dict
            Dictionary assigning each row and column of the laplacian matrix to a
            simplex. Only returned if `index` is True.
        down_laplacian : scipy.sparse.csr.csr_matrix
            The down laplacian matrix of this simplicial complex.
        """
        if weight is not None:
            raise ValueError("`weight` is not supported in this version")

        if self.dim >= rank > 0:
            _, column, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_down = B.transpose() @ B
        else:
            raise ValueError(
                f"Rank should be larger than 1 and <= {self.dim} (maximal dimension cells), got {rank}."
            )
        if not signed:
            L_down = abs(L_down)
        if index:
            return column, L_down
        return L_down

    def adjacency_matrix(
        self, rank, signed: bool = False, weight: str | None = None, index: bool = False
    ):
        """Compute the adjacency matrix of the simplicial complex.

        Parameters
        ----------
        rank : int
            Rank of the adjacency matrix.
        signed : bool
            Whether to return the signed or unsigned adjacency matrix.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the adjacency
            matrix.

        Returns
        -------
        indices : dict
            Dictionary mapping each row and column of the coadjacency matrix to a
            simplex. Only returned if `index` is True.
        adjacency_matrix : scipy.sparse.csr.csr_matrix
            The adjacency matrix of rank `rank` of this simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([1, 2, 3, 4])
        >>> SC.add_simplex([1, 2, 4])
        >>> SC.add_simplex([3, 4, 8])
        >>> adj1 = SC.adjacency_matrix(1)
        """
        if weight is not None:
            raise ValueError("`weight` is not supported in this version")

        ind, L_up = self.up_laplacian_matrix(
            rank, signed=signed, weight=weight, index=True
        )
        L_up.setdiag(0)

        if not signed:
            L_up = abs(L_up)
        if index:
            return ind, L_up
        return L_up

    def coadjacency_matrix(
        self, rank: int, signed: bool = False, weight=None, index: bool = False
    ):
        """Compute the coadjacency matrix of the simplicial complex.

        Parameters
        ----------
        rank : int
            Rank of the coadjacency matrix.
        signed : bool
            Whether to return the signed or unsigned coadjacency matrix.
        weight : str, optional
            The name of the simplex attribute to use as weights for the hodge laplacian
            matrix. If `None`, the matrix is binary.
        index : bool, default=False
            Whether to return the indices of the rows and columns of the coadjacency
            matrix.

        Returns
        -------
        indices : list
            List identifying the rows and columns of the coadjacency matrix with the
            simplices of the simplicial complex. Only returned if `index` is True.
        coadjacency_matrix : scipy.sparse.csr.csr_matrix
            The coadjacency matrix of this simplicial complex.
        """
        if weight is not None:
            raise ValueError("`weight` is not supported in this version")

        ind, L_down = self.down_laplacian_matrix(
            rank, signed=signed, weight=weight, index=True
        )
        L_down.setdiag(0)
        if not signed:
            L_down = abs(L_down)
        if index:
            return ind, L_down
        return L_down

    def add_elements_from_nx_graph(self, G: nx.Graph) -> None:
        """Add elements from a networkx graph to this simplicial complex.

        Parameters
        ----------
        G : networkx.Graph
            A networkx graph instance.
        """
        _simplices = [[n] for n in G.nodes] + list(G.edges)
        self.add_simplices_from(_simplices)

    def restrict_to_simplices(self, cell_set) -> Self:
        """Construct a simplicial complex using a subset of the simplices.

        Parameters
        ----------
        cell_set : iterable of hashables or simplices
            A subset of elements of the simplicial complex that should be in the new
            simplicial complex.

        Returns
        -------
        SimplicialComplex
            New simplicial complex restricted to the simplices in `cell_set`.

        Examples
        --------
        >>> c1 = tnx.Simplex((1, 2, 3))
        >>> c2 = tnx.Simplex((1, 2, 4))
        >>> c3 = tnx.Simplex((1, 2, 5))
        >>> SC = tnx.SimplicialComplex([c1, c2, c3])
        >>> SC1 = SC.restrict_to_simplices([c1, (2, 4)])
        >>> SC1.simplices
        SimplexView([(1,), (2,), (3,), (4,), (1, 2), (1, 3), (2, 3), (2, 4), (1, 2, 3)])
        """
        rns = [cell for cell in cell_set if cell in self.simplices]
        return self.__class__(simplices=rns)

    def restrict_to_nodes(self, node_set) -> "SimplicialComplex[ElementType]":
        """Construct a new simplicial complex by restricting the simplices.

        The simplices are restricted to the nodes referenced by node_set.

        Parameters
        ----------
        node_set : iterable of hashables
            A subset of nodes of the simplicial complex to restrict to.

        Returns
        -------
        SimplicialComplex
            A new simplicial complex restricted to the nodes in `node_set`.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([(1, 2, 3), (1, 2, 4), (1, 2, 5)])
        >>> new_complex = SC.restrict_to_nodes([1, 2, 3, 4])
        >>> new_complex.simplices
        SimplexView([(1,), (2,), (3,), (4,), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (1, 2, 3), (1, 2, 4)])
        """
        node_set = set(node_set)
        simplices = [
            s
            for rank in range(1, self.dim + 1)
            for s in self.skeleton(rank)
            if set(s).issubset(node_set)
        ]
        all_sim = simplices + [frozenset({i}) for i in node_set if i in self.nodes]

        return SimplicialComplex(all_sim)

    def get_all_maximal_simplices(self) -> Generator[Simplex[ElementType], None, None]:
        """Get all maximal simplices of this simplicial complex.

        A simplex is maximal if it is not a face of any other simplex in the complex.

        Returns
        -------
        list of tuples
            List of maximal simplices in this simplicial complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex([(1, 2), (1, 2, 3), (1, 2, 4), (2, 5)])
        >>> SC.get_all_maximal_simplices()
        [(2, 5), (1, 2, 3), (1, 2, 4)]
        """
        for node in self._simplex_trie:
            simplex = node.simplex
            if self.is_maximal(simplex):
                yield simplex

    @classmethod
    def from_spharapy(cls, mesh) -> Self:
        """Import from sharpy.

        Parameters
        ----------
        mesh : spharapy.trimesh.TriMesh
            The input spharapy object.

        Returns
        -------
        SimplicialComplex
            The resulting SimplicialComplex.

        Examples
        --------
        >>> import spharapy.trimesh as tm
        >>> import spharapy.spharabasis as sb
        >>> import spharapy.datasets as sd
        >>> mesh = tm.TriMesh(
        ...     [[0, 1, 2]], [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]]
        ... )

        >>> SC = tnx.SimplicialComplex.from_spharapy(mesh)
        """
        vertices = np.array(mesh.vertlist)
        SC = cls(mesh.trilist)

        first_ind = np.min(mesh.trilist)

        if first_ind == 0:
            SC.set_simplex_attributes(dict(enumerate(vertices)), name="position")
        else:  # first index starts at 1.
            SC.set_simplex_attributes(
                dict(enumerate(vertices, first_ind)),
                name="position",
            )

        return SC

    def to_hasse_graph(self) -> nx.DiGraph:
        """Create the hasse graph corresponding to this simplicial complex.

        Returns
        -------
        nx.DiGraph
            A NetworkX Digraph representing the Hasse graph of the Simplicial Complex.

        Examples
        --------
        >>> SC = tnx.SimplicialComplex()
        >>> SC.add_simplex([0, 1, 2])
        >>> G = SC.to_hasse_graph()
        """
        G = nx.DiGraph()
        for n in self.nodes:
            G.add_node(tuple(sorted(n)))
        for i in range(1, self.dim + 1):
            for c in self.skeleton(i):
                G.add_node(tuple(sorted(c)))
                for f in combinations(c, len(c) - 1):
                    G.add_edge(tuple(sorted(f)), tuple(sorted(c)))
        return G

    @classmethod
    def from_gudhi(cls, tree: SimplexTree) -> Self:
        """Import from gudhi.

        Parameters
        ----------
        tree : gudhi.SimplexTree
            The input gudhi simplex tree.

        Returns
        -------
        SimplicialComplex
            The resulting SimplicialComplex.

        Examples
        --------
        >>> from gudhi import SimplexTree
        >>> tree = SimplexTree()
        >>> _ = tree.insert([1, 2, 3, 5])
        >>> SC = tnx.SimplicialComplex.from_gudhi(tree)
        """
        SC = cls()
        for simplex, _ in tree.get_skeleton(tree.dimension()):
            SC.add_simplex(simplex)
        return SC

    @classmethod
    def from_trimesh(cls, mesh) -> Self:
        """Import from trimesh.

        Parameters
        ----------
        mesh : trimesh.Trimesh
            The input trimesh object.

        Returns
        -------
        SimplicialComplex
            The resulting SimplicialComplex.

        Examples
        --------
        >>> import trimesh
        >>> mesh = trimesh.Trimesh(
        ...     vertices=[[0, 0, 0], [0, 0, 1], [0, 1, 0]],
        ...     faces=[[0, 1, 2]],
        ...     process=False,
        ... )
        >>> SC = tnx.SimplicialComplex.from_trimesh(mesh)
        >>> print(SC.nodes)
        >>> print(SC.simplices)
        >>> SC[(0)]["position"]
        """
        SC = cls(mesh.faces)

        first_ind = np.min(mesh.faces)

        if first_ind == 0:
            SC.set_simplex_attributes(
                dict(enumerate(mesh.vertices)),
                name="position",
            )
        else:  # first index starts at 1.
            SC.set_simplex_attributes(
                dict(enumerate(mesh.vertices, first_ind)),
                name="position",
            )

        return SC

    @classmethod
    def load_mesh(cls, file_path, process: bool = False) -> Self:
        """Load a mesh.

        Parameters
        ----------
        file_path : str or pathlib.Path
            The source of the data to be loaded.
        process : bool
            Whether trimesh should try to process the mesh before loading it.

        Returns
        -------
        SimplicialComplex
            The output simplicial complex stores the same structure stored in the mesh input file.

        Notes
        -----
        mesh files supported : obj, off, glb

        Examples
        --------
        >>> SC = tnx.SimplicialComplex.load_mesh("C:/temp/stanford-bunny.obj")
        >>> SC.nodes
        """
        import trimesh

        mesh = trimesh.load_mesh(file_path, process=process, force=None)
        return cls.from_trimesh(mesh)

    def is_triangular_mesh(self) -> bool:
        """Check if the simplicial complex is a triangular mesh.

        Returns
        -------
        bool
            True if the simplicial complex can be converted to a triangular mesh,
            False otherwise.
        """
        if self.dim > 2:
            return False

        lst = self.get_all_maximal_simplices()
        return all(len(i) != 2 for i in lst)

    def to_trimesh(self, vertex_position_name: str = "position"):
        """Convert simplicial complex to trimesh object.

        Parameters
        ----------
        vertex_position_name : str, default="position"
            The simplex attribute name that contains the vertex positions.

        Returns
        -------
        trimesh.Trimesh
            The trimesh object corresponding to this simplicial complex.
        """
        import trimesh

        if not self.is_triangular_mesh():
            raise RuntimeError(
                "input simplicial complex has dimension higher than 2 and hence it cannot be converted to a trimesh object"
            )

        vertices = np.array(
            list(
                dict(
                    sorted(self.get_node_attributes(vertex_position_name).items())
                ).values()
            )
        )
        faces = [simplex.elements for simplex in self.get_all_maximal_simplices()]

        return trimesh.Trimesh(faces=faces, vertices=vertices, process=False)

    def to_spharapy(self, vertex_position_name: str = "position"):
        """Convert to spharapy.

        Parameters
        ----------
        vertex_position_name : str, default="position"
            The simplex attribute name that contains the vertex positions.

        Returns
        -------
        spharapy.trimesh.TriMesh
            The spharapy object corresponding to this simplicial complex.

        Raises
        ------
        RuntimeError
            If package `spharapy` is not installed.

        Examples
        --------
        >>> import spharapy.trimesh as tm
        >>> import spharapy.spharabasis as sb
        >>> import spharapy.datasets as sd
        >>> mesh = tm.TriMesh([[0, 1, 2]], [[0, 0, 0], [0, 0, 1], [0, 1, 0]])
        >>> SC = tnx.SimplicialComplex.from_spharapy(mesh)
        >>> mesh2 = SC.to_spharapy()
        >>> mesh2.vertlist == mesh.vertlist
        >>> mesh2.trilist == mesh.trilist
        """
        try:
            import spharapy.trimesh as tm
        except ImportError as e:
            raise RuntimeError(
                "Package `spharapy` is required for this function."
            ) from e

        if not self.is_triangular_mesh():
            raise RuntimeError(
                "Simplicial complex has dimension higher than 2 and cannot be converted to a trimesh object."
            )

        vertices = list(
            dict(
                sorted(self.get_node_attributes(vertex_position_name).items())
            ).values()
        )
        triangles = [simplex.elements for simplex in self.get_all_maximal_simplices()]

        return tm.TriMesh(triangles, vertices)

    def laplace_beltrami_operator(self, mode: str = "inv_euclidean"):
        """Compute a laplacian matrix for a triangular mesh.

        The method creates a laplacian matrix for a triangular
        mesh using different weighting function.

        Parameters
        ----------
        mode : {'unit', 'inv_euclidean', 'half_cotangent'}, optional
            The methods for determining the edge weights. Using the option
            'unit' all edges of the mesh are weighted by unit weighting
            function, the result is an adjacency matrix. The option
            'inv_euclidean' results in edge weights corresponding to the
            inverse Euclidean distance of the edge lengths. The option
            'half_cotangent' uses the half of the cotangent of the two angles
            opposed to an edge as weighting function. the default weighting
            function is 'inv_euclidean'.

        Returns
        -------
        numpy.ndarray, shape (n_vertices, n_vertices)
            Matrix, which contains the discrete laplace operator for data
            defined at the vertices of a triangular mesh. The number of
            vertices of the triangular mesh is n_points.
        """
        mesh = self.to_spharapy()
        return mesh.laplacianmatrix(mode=mode)

    @classmethod
    def from_nx(cls, G: nx.Graph) -> Self:
        """Convert from netwrokx graph.

        Parameters
        ----------
        G : nx.Graph
            The graph to convert to a simplicial complex.

        Returns
        -------
        SimplicialComplex
            The simplicial complex.

        Examples
        --------
        >>> G = nx.Graph()  # or DiGraph, MultiGraph, MultiDiGraph, etc
        >>> G.add_edge(1, 2, weight=2)
        >>> G.add_edge(3, 4, weight=4)
        >>> SC = tnx.SimplicialComplex.from_nx(G)
        >>> SC[(1, 2)]["weight"]
        2
        """
        return cls(G)

    def is_connected(self) -> bool:
        """Check if the simplicial complex is connected.

        Returns
        -------
        bool
            True if the simplicial complex is connected, False otherwise.

        Notes
        -----
        A simplicial complex is connected iff its 1-skeleton is connected.
        """
        return nx.is_connected(self.graph_skeleton())

    @classmethod
    def simplicial_closure_of_hypergraph(cls, H) -> Self:
        """Compute the simplicial complex closure of a hypergraph.

        Parameters
        ----------
        H : hyernetx hypergraph
            The hypergraph to compute the simplicial complex closure of.

        Returns
        -------
        SimplicialComplex
            Simplicial complex closure of the hypergraph.

        Examples
        --------
        >>> import hypernetx as hnx
        >>> hg = hnx.Hypergraph([[1, 2, 3, 4], [1, 2, 3]], static=True)
        >>> sc = tnx.SimplicialComplex.simplicial_closure_of_hypergraph(hg)
        >>> sc.simplices
        SimplexView([(1,), (2,), (3,), (4,), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4), (1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4), (1, 2, 3, 4)])
        """
        edges = H.edges
        lst = [edges[e] for e in edges]
        return cls(lst)

    def to_cell_complex(self):
        """Convert a simplicial complex to a cell complex.

        Returns
        -------
        toponetx.classes.CellComplex
            The cell complex corresponding to this simplicial complex.

        Examples
        --------
        >>> c1 = tnx.Simplex((1, 2, 3))
        >>> c2 = tnx.Simplex((1, 2, 4))
        >>> c3 = tnx.Simplex((2, 5))
        >>> SC = tnx.SimplicialComplex([c1, c2, c3])
        >>> SC.to_cell_complex()
        """
        from toponetx.classes.cell_complex import CellComplex

        return CellComplex(
            simplex.elements for simplex in self.get_all_maximal_simplices()
        )

    def to_hypergraph(self):
        """Convert a simplicial complex to a `hypernetx` hypergraph.

        Returns
        -------
        Hypergraph
            The hypergraph corresponding to this simplicial complex.

        Raises
        ------
        RuntimeError
            If `hypernetx` is not installed

        Examples
        --------
        >>> c1 = tnx.Simplex((1, 2, 3))
        >>> c2 = tnx.Simplex((1, 2, 4))
        >>> c3 = tnx.Simplex((2, 5))
        >>> SC = tnx.SimplicialComplex([c1, c2, c3])
        >>> SC.to_hypergraph()
        Hypergraph({'e0': [1, 2], 'e1': [1, 3], 'e2': [1, 4], 'e3': [2, 3], 'e4': [2, 4], 'e5': [2, 5], 'e6': [1, 2, 3], 'e7': [1, 2, 4]},name=)
        """
        if Hypergraph is None:
            raise RuntimeError(
                "Cannot transform simplicial complex to hypergraph, `hypernetx` is not installed."
            )

        edges = {}
        edge_properties = {}

        for simplex in filter(lambda s: len(s) > 1, self.simplices):
            edges[str(list(simplex))] = simplex
            edge_properties[str(list(simplex))] = self[simplex]

        HG = Hypergraph(edges, edge_properties=edge_properties)
        HG.add_nodes_from([(next(iter(node)), self[node]) for node in self.nodes])

        return HG

    def to_combinatorial_complex(self):
        """Convert a simplicial complex to a combinatorial complex.

        Returns
        -------
        toponetx.classes.CombinatorialComplex
            The combinatorial complex equivalent to this simplicial complex.

        Examples
        --------
        >>> c1 = tnx.Simplex((1, 2, 3))
        >>> c2 = tnx.Simplex((1, 2, 3))
        >>> c3 = tnx.Simplex((1, 2, 4))
        >>> SC = tnx.SimplicialComplex([c1, c2, c3])
        >>> CCC = SC.to_combinatorial_complex()
        """
        from toponetx.classes.combinatorial_complex import CombinatorialComplex

        CCC = CombinatorialComplex()
        for rank in range(1, self.dim + 1):
            for cell in self.skeleton(rank):
                CCC.add_cell(cell, rank=len(cell) - 1, **self[cell])
        return CCC

    def graph_skeleton(self) -> nx.Graph:
        """Return the graph-skeleton of this simplicial complex.

        The graph-skeleton consists of the 0 and 1-simplices (i.e., nodes and edges)
        of the simplicial complex.

        Returns
        -------
        nx.Graph
            The graph-skeleton of this simplicial complex.
        """
        G = nx.Graph()
        for node in self.skeleton(0):
            G.add_node(next(iter(node)), **self[node])
        for edge in self.skeleton(1):
            G.add_edge(*edge, **self[edge])
        return G

    def clone(self) -> Self:
        """Return a copy of the simplicial complex.

        The clone method by default returns an independent shallow copy of the
        simplicial complex. Use Python's `copy.deepcopy` for new containers.

        Returns
        -------
        SimplicialComplex
            A shallow copy of this simplicial complex.
        """
        return self.__class__(self.simplices)



"""Methods to lift a graph to a cell complex."""

__all__ = [
    "homology_cycle_cell_complex",
]


import networkx as nx

from toponetx.classes.cell_complex import CellComplex


def homology_cycle_cell_complex(G) -> CellComplex:
    """Get the cell complex obtained by adding homology cycles of a graph.

    Parameters
    ----------
    G : networkx graph
        Input graph.

    Returns
    -------
    CellComplex
        The cell complex obtained by adding the homology cycles of the graph.
    """
    cycles = nx.cycle_basis(G)
    cx = CellComplex(G)
    cx.add_cells_from(cycles, rank=2)
    return cx
